#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# Copyright 2008-2016 Glencoe Software, Inc.  All Rights Reserved.
# Use is subject to license terms supplied in LICENSE.txt
# Josh Moore, josh at glencoesoftware.com
"""
OMERO Grid admin controller.

This is a python wrapper around icegridregistry/icegridnode for master
and various other tools needed for administration.
"""
from __future__ import division
from __future__ import print_function

from builtins import str
from future.utils import bytes_to_native_str
from future.utils import isbytes
from past.utils import old_div
from builtins import object
import re
import os
import sys
import platform
import datetime
import time

from glob import glob
from math import ceil
from zipfile import ZipFile

import omero
import omero.config

from omero.grid import RawAccessRequest

from omero.cli import (
    admin_only,
    CLI,
    DiagnosticsControl,
    DirectoryType,
    NonZeroReturnCode,
    ServiceManagerMixin,
    UserGroupControl,
)

from omero.install.config_parser import PropertyParser
from omero.plugins.prefs import \
    WriteableConfigControl, with_config
from omero.install.windows_warning import windows_warning, WINDOWS_WARNING

import portalocker
from omero_ext.path import path
from omero_ext.which import whichall
from omero_ext.argparse import FileType
from omero_version import ice_compatibility

from omero.util._process_defaultxml import _process_xml

try:
    from omero_version import omero_version
    VERSION = omero_version
except ImportError:
    VERSION = "Unknown"  # Usually during testing

try:
    import pywintypes
    import win32service
    import win32evtlogutil
    import win32api
    import win32security
    has_win32 = True
except ImportError:
    has_win32 = False

DEFAULT_WAIT = 300

CHECKUPGRADE_USERAGENT = "test"

HELP = """Administrative tools including starting/stopping OMERO.

Environment variables:
 OMERO_MASTER
 OMERO_NODE

Configuration properties:
 omero.windows.user
 omero.windows.pass
 omero.windows.servicename
 omero.web.application_server.port
 omero.web.server_list

""" + "\n" + "="*50 + "\n"


if platform.system() == 'Windows':
    HELP += ("\n\n%s" % WINDOWS_WARNING)


class AdminControl(DiagnosticsControl,
                   WriteableConfigControl,
                   UserGroupControl,
                   ServiceManagerMixin):

    # Require by omero.cli.ServiceManagerMixin
    SERVICE_MANAGER_KEY = 'server'

    def _complete(self, text, line, begidx, endidx):
        """
        Returns a file after "deploy", "start", or "startasync"
        and otherwise delegates to the BaseControl
        """
        for s in (" deploy ", " start ", " startasync "):
            length = len(s)
            i = line.find(s)
            if i >= 0:
                f = line[i+length:]
                return self._complete_file(f)
        return WriteableConfigControl._complete(
            self, text, line, begidx, endidx)

    def _configure(self, parser):
        sub = parser.sub()
        parser = self._add_diagnostics(parser, sub)
        parser.add_argument(
            "--all-jars", action="store_true",
            help="Show information for all jars")
        self.add_error(
            "NOT_WINDOWS", 123,
            "Not Windows")
        self.add_error(
            "SETUP", 200,
            "Error during service user set up:  (%s) %s")
        self.add_error(
            "RUNNING", 201,
            "%s is already running. Use stop first")
        self.add_error(
            "NO_SERVICE", 202,
            "%s service deleted.")
        self.add_error(
            "BAD_CONFIG", 300,
            "Bad configuration: No IceGrid.Node.Data property")
        self.add_error(
            "WIN_CONFIG", 400, r"""

            %s is not in this directory. Aborting...

            Please see the installation instructions on modifying
            the files for your installation (%s)
            with bin\winconfig.bat

            """)
        self.add_error(
            "NO_WIN32", 666,
            "Could not import win32service and/or win32evtlogutil")
        self.actions = {}

        class Action(object):
            def __init__(this, name, help, wait=False):
                this.parser = sub.add_parser(name, help=help,
                                             description=help)
                this.parser.set_defaults(func=getattr(self, name))
                self.actions[name] = this.parser
                if wait:
                    this.parser.add_argument(
                        "--wait", type=float, default=DEFAULT_WAIT,
                        help="Seconds to wait for operation")

        Action(
            "start",
            """Start icegridnode daemon and waits for required components to \
come up, i.e. status == 0

If the first argument can be found as a file, it will be deployed as the
application descriptor rather than etc/grid/default.xml. All other arguments
will be used as targets to enable optional sections of the descriptor.

On Windows, two arguments (-u and -w) specify the Windows service Log On As
user credentials. If not specified, omero.windows.user and omero.windows.pass
will be used.""",
            wait=True)

        Action("startasync", "The same as start but returns immediately",)

        Action("restart", "stop && start", wait=True)

        Action(
            "restartasync", """The same as restart but returns as soon as \
starting has begun.""",
            wait=True)

        Action("status", """Status of server

Returns with 0 status if a node ping is successful and if some SessionManager
returns an OMERO-specific exception on a bad login. This can be used in shell
scripts, e.g.:

    $ omero admin status && echo "server started"
            """)

        Action(
            "stop",
            """Initiates node shutdown and waits for status to return a \
non-0 value""",
            wait=True)

        Action("stopasync", "The same as stop but returns immediately")

        Action(
            "deploy",
            """Deploy the given deployment descriptor. See etc/grid/*.xml

If the first argument is not a file path, etc/grid/default.xml will be
deployed by default. Same functionality as start, but requires that the node
already be running. This may automatically restart some server components.""")

        Action(
            "ice", "Drop user into icegridadmin console or execute arguments")

        fixpyramids = Action(
            "fixpyramids",
            "Remove empty pyramid pixels files (admins only)").parser

        removepyramids = Action(
            "removepyramids",
            """Remove pyramid pixels files (admins only) according to endianness.
By default pyramids with big and little-endianness will be deleted.
To delete pyramids with little-endianness equals to true use --endian=little.

Examples:
  omero admin removepyramids --dry-run
  omero admin removepyramids --dry-run --endian=little
  omero admin removepyramids --dry-run --imported-after YYYY-mm-dd
            """).parser
        # See cleanse options below

        email = Action(
            "email",
            """Send administrative emails to users.

Administrators can contact OMERO users and groups of
users based on configured email settings. A subject
and some text are required. If no text is passed on
the command-line or if "-" is passed, then text will
be read from the standard input.

Examples:

  # Send the contents of a file to everyone
  # except inactive users.
  omero admin email --everyone Subject < some_file.text

  # Include inactive users in the email
  omero admin email --everyone --inactive ...

  # Contact a single group
  omero admin email --group-name system \\
                        Subject short message

  # Contact a list of users
  omero admin email --user-id 10 \\
                        --user-name ralph \\
                        Subject ...

            """).parser
        email.add_argument(
            "subject",
            help="Required subject for the mail")
        email.add_argument(
            "text", nargs="*",
            help=("All further arguments are combined "
                  "to form the body. stdin if none or '-' "
                  "is given."))
        email.add_argument(
            "--everyone", action="store_true",
            help=("Contact everyone in the system regardless "
                  "of other arguments."))
        email.add_argument(
            "--inactive", action="store_true",
            help="Do not filter inactive users.")
        self._add_wait(email)
        self.add_user_and_group_arguments(email,
                                          action="append",
                                          exclusive=False)

        Action(
            "rewrite",
            """Regenerate the IceGrid configuration files

Regenerates the IceGrid configuration files from the template files and the
current configuration properties. Recalculates the JVM configuration settings
and replaces the memory settings as well as the port and host properties under
the corresponding application descriptors.
            """)

        Action(
            "jvmcfg",
            "Display JVM configuration settings based on the current system")

        Action(
            "waitup",
            "Used by start after calling startasync to wait on status==0",
            wait=True)

        Action(
            "waitdown",
            "Used by stop after calling stopasync to wait on status!=0",
            wait=True)

        reindex = Action(
            "reindex",
            """Re-index the Lucene index

Command-line tool for re-indexing the database. This command must be run on the
machine where the FullText directory is located. In most cases, you will want
to disable the background indexer before running most of these commands.

See https://docs.openmicroscopy.org/latest/omero/sysadmins/search.html
for more information.

Examples:

  # 1. Reset the 'last indexed' counter. Defaults to 0
  omero admin reindex --reset 0

  # 2. Delete the contents of a corrupt FullText directory
  omero admin reindex --wipe

  # 3. Run indexer in the foreground. Disable the background first
  omero admin reindex --foreground
  # Foreground indexing is NOT currently working.

Other commands (usually unnecessary):

  # Index all objects in the database.
  omero admin reindex --full

  # Index one specific class
  omero admin reindex --class ome.model.core.Image

  # Passing arguments to Java
  JAVA_OPTS="-Dlogback.configurationFile=stderr.xml" \\
  omero admin reindex --foreground

  JAVA_OPTS="-Xdebug -Xrunjdwp:server=y,transport=\
dt_socket,address=8787,suspend=y" \\
  omero admin reindex --foreground

""").parser

        reindex.add_argument(
            "--jdwp", action="store_true",
            help="Activate remote debugging")
        reindex.add_argument(
            "--mem", default="1024m",
            help="Heap size to use")
        reindex.add_argument(
            "--batch", default="500",
            help="Number of items to index before reporting status")
        reindex.add_argument(
            "--merge-factor", "--merge_factor", default="100",
            help=("Higher means merge less frequently. "
                  "Faster but needs more RAM"))
        reindex.add_argument(
            "--ram-buffer-size", "--ram_buffer_size", default="1000",
            help=("Number of MBs to use for the indexing. "
                  "Higher is faster."))
        reindex.add_argument(
            "--lock-factory", "--lock_factory", default="native",
            help=("Choose Lucene lock factory by class or "
                  "'native', 'simple', 'none'"))

        group = reindex.add_mutually_exclusive_group()
        group.add_argument(
            "--prepare", action="store_true",
            help="Disables the background indexer in preparation for indexing")
        group.add_argument(
            "--full", action="store_true",
            help="Reindexes all non-excluded tables sequentially")
        group.add_argument(
            "--events", action="store_true",
            help="Reindexes all non-excluded event logs chronologically")
        group.add_argument(
            "--reset", default=None,
            help="Reset the index counter")
        group.add_argument(
            "--dryrun", action="store_true",
            help=("Run through all events, incrementing the counter. "
                  "NO INDEXING OCCURS"))
        group.add_argument(
            "--foreground", action="store_true",
            help=("Run indexer in the foreground"))
        group.add_argument(
            "--class", nargs="+",
            help="Reindexes the given classes sequentially")
        group.add_argument(
            "--wipe", action="store_true",
            help="Delete the existing index files")
        group.add_argument(
            "--finish", action="store_true",
            help="Re-enables the background indexer after for indexing")

        sessionlist = Action(
            "sessionlist", "List currently running sessions (deprecated)") \
            .parser
        sessionlist.add_login_arguments()

        cleanse = Action("cleanse", """Remove binary data files from OMERO  (admins only)

Deleting an object from OMERO currently may not remove all the binary data.
Use this command either manually or in a cron job periodically to remove
Pixels, empty directories, and other data.

This is done by checking that for all the files in the given directory, a
matching entry exists on the server. THE /OMERO DIRECTORY MUST MATCH THE
DATABASE YOU ARE RUNNING AGAINST.

This command must be run on the machine where, for example, /OMERO/ is
located.

Examples:
  omero admin cleanse --dry-run /OMERO      # Lists files that will be \
deleted
  omero admin cleanse /OMERO                # Actually delete them.
  omero admin cleanse /volumes/data/OMERO   # Delete from a standard \
location.

""").parser

        for x in (cleanse, fixpyramids):
            x.add_argument(
                "--dry-run", action="store_true",
                help="Print out which files would be deleted")
            x.add_argument(
                "data_dir", type=DirectoryType(),
                help="omero.data.dir directory value e.g. /OMERO")
            x.add_login_arguments()

        removepyramids.add_argument(
            "--dry-run", action="store_true",
            help="Print out which files would be deleted")
        removepyramids.add_argument(
            "--endian", choices=("little", "big", "both"), default="both",
            help="Delete pyramid with given endianness. "
            "If not specified, all will be removed.")
        removepyramids.add_argument(
            "--imported-after", metavar="DATE",
            help="Delete pyramid imported after a given date. "
            "Expected format YYYY-mm-dd")
        removepyramids.add_argument(
            "--limit", metavar="MAX_NUMBER",
            help="Set the limit of pyramids to remove in one call. "
            "Values greater than 500 (default) are not supported")
        removepyramids.add_login_arguments()
        self._add_wait(removepyramids)

        Action("checkwindows", "Run simple check of the local installation "
               "(Windows-only)")
        Action("checkice", "Run simple check of the Ice installation")

        Action("events", "Print event log (Windows-only)")

        Action(
            "checkupgrade", "Check whether a server upgrade is available")

        log = Action("log", "Add a custom log message to "
                            "the server log").parser
        log.add_argument(
            "--level",
            help="The log level: trace, debug, info, warn or error "
                 "(default: info)", default="info")
        log.add_argument(
            "repo",
            help="The repo uuid (e.g. ScriptRepo)")
        log.add_argument(
            "message",
            help="The log message to add")
        log.add_login_arguments()

        self.actions["ice"].add_argument(
            "argument", nargs="*",
            help="""Arguments joined together to make an Ice command. If not \
present, the user will enter a console""")

        self.actions["status"].add_argument(
            "node", nargs="?", default="master")
        self.actions["status"].add_argument(
            "--nodeonly", action="store_true",
            help="If set, then only tests if the icegridnode is running")

        for name in ("start", "startasync", "restart", "restartasync", "stop",
                     "stopasync"):
            self.actions[name].add_argument(
                "--force-rewrite", action="store_true",
                help="Force the configuration to be rewritten before checking"
                " the server status")

        for name in ("start", "restart"):
            self.actions[name].add_argument(
                "--foreground", action="store_true",
                help="Start server in foreground mode (no daemon/service)")

        for name in ("start", "startasync", "restart", "restartasync"):
            self.actions[name].add_argument(
                "-u", "--user",
                help="Windows Service Log On As user name.")
            self.actions[name].add_argument(
                "-w", "--password", metavar="PW",
                help="Windows Service Log On As user password.")

        for name in ("start", "startasync", "deploy", "restart",
                     "restartasync"):
            self.actions[name].add_argument(
                "file", nargs="?",
                help="Application descriptor. If not provided, a default"
                " will be used")
            self.actions[name].add_argument(
                "targets", nargs="*",
                help="Targets within the application descriptor which "
                " should  be activated. Common values are: \"debug\", "
                "\"trace\" ")

        # DISABLED = """ see: http://www.zeroc.com/forums/bug-reports/\
        # 4237-sporadic-freeze-errors-concurrent-icegridnode-access.html
        #   restart [filename] [targets]      : Calls stop followed by start \
        #   args
        #   restartasync [filename] [targets] : Calls stop followed by \
        #   startasync args
        # """

    #
    # Windows utility methods
    #
    if has_win32:
        def _get_service_name(unused, config):
            try:
                return config.as_map()["omero.windows.servicename"]
            except KeyError:
                return 'OMERO'

        def _query_service(unused, svc_name):
            hscm = win32service.OpenSCManager(
                None, None, win32service.SC_MANAGER_ALL_ACCESS)
            try:
                try:
                    hs = win32service.OpenService(
                        hscm, svc_name, win32service.SERVICE_ALL_ACCESS)
                except Exception:
                    return "DOESNOTEXIST"
                try:
                    q = win32service.QueryServiceStatus(hs)
                    type, state, ctrl, err, svcerr, svccp, svcwh = q
                    if state == win32service.SERVICE_STOPPED:
                        return "STOPPED"
                    else:
                        return "unknown"
                finally:
                    win32service.CloseServiceHandle(hs)
            finally:
                win32service.CloseServiceHandle(hscm)

        def _stop_service(self, svc_name):
            hscm = win32service.OpenSCManager(
                None, None, win32service.SC_MANAGER_ALL_ACCESS)
            try:
                hs = win32service.OpenService(
                    hscm, svc_name, win32service.SC_MANAGER_ALL_ACCESS)
                win32service.ControlService(
                    hs, win32service.SERVICE_CONTROL_STOP)
                win32service.DeleteService(hs)
                self.ctx.out("%s service deleted." % svc_name)
            finally:
                win32service.CloseServiceHandle(hs)
                win32service.CloseServiceHandle(hscm)

        def _start_service(self, config, descript, svc_name, pasw, user):
            output = self._query_service(svc_name)
            # Now check if the server exists
            if 0 <= output.find("DOESNOTEXIST"):
                binpath = """icegridnode.exe "%s" --deploy "%s" --service\
                    %s""" % (self._icecfg(), descript, svc_name)

                # By default: "NT Authority\Local System"
                if not user:
                    try:
                        user = config.as_map()["omero.windows.user"]
                    except KeyError:
                        user = None
                if user is not None and len(user) > 0:
                    if "\\" not in user:
                        computername = win32api.GetComputerName()
                        user = "\\".join([computername, user])
                    try:
                        # See #9967, code based on http://mail.python.org/\
                        # pipermail/python-win32/2010-October/010791.html
                        self.ctx.out("Granting SeServiceLogonRight to service"
                                     " user \"%s\"" % user)
                        policy_handle = win32security.LsaOpenPolicy(
                            None, win32security.POLICY_ALL_ACCESS)
                        sid_obj, domain, tmp = \
                            win32security.LookupAccountName(None, user)
                        win32security.LsaAddAccountRights(
                            policy_handle, sid_obj, ('SeServiceLogonRight',))
                        win32security.LsaClose(policy_handle)
                    except pywintypes.error as details:
                        self.raise_error("SETUP", details[0], details[2])
                    if not pasw:
                        try:
                            pasw = config.as_map()["omero.windows.pass"]
                        except KeyError:
                            pasw = self._ask_for_password(
                                " for service user \"%s\"" % user)
                else:
                    pasw = None

                hscm = win32service.OpenSCManager(
                    None, None, win32service.SC_MANAGER_ALL_ACCESS)
                try:
                    self.ctx.out("Installing %s Windows service." % svc_name)
                    hs = win32service.CreateService(
                        hscm, svc_name, svc_name,
                        win32service.SERVICE_ALL_ACCESS,
                        win32service.SERVICE_WIN32_OWN_PROCESS,
                        win32service.SERVICE_AUTO_START,
                        win32service.SERVICE_ERROR_NORMAL, binpath, None, 0,
                        None, user, pasw)
                    self.ctx.out("Successfully installed %s Windows service."
                                 % svc_name)
                    win32service.CloseServiceHandle(hs)
                finally:
                    win32service.CloseServiceHandle(hscm)

            # Then check if the server is already running
            if 0 <= output.find("RUNNING"):
                self.raise_error("RUNNING", svc_name)

            # Finally, try to start the service - delete if startup fails
            hscm = win32service.OpenSCManager(
                None, None, win32service.SC_MANAGER_ALL_ACCESS)
            try:
                try:
                    hs = win32service.OpenService(
                        hscm, svc_name, win32service.SC_MANAGER_ALL_ACCESS)
                    win32service.StartService(hs, None)
                    self.ctx.out("Starting %s Windows service." % svc_name)
                except pywintypes.error as details:
                    self.ctx.out("%s service startup failed: (%s) %s"
                                 % (svc_name, details[0], details[2]))
                    win32service.DeleteService(hs)
                    self.raise_error("NO_SERVICE", svc_name)
            finally:
                win32service.CloseServiceHandle(hs)
                win32service.CloseServiceHandle(hscm)

        def events(self, svc_name):
            def DumpRecord(record):
                if str(record.SourceName) == svc_name:
                    self.ctx.out("Time: %s" % record.TimeWritten)
                    self.ctx.out("Rec:  %s" % record.RecordNumber)
                    for si in record.StringInserts:
                        self.ctx.out(si)
                    self.ctx.out("="*20)
            win32evtlogutil.FeedEventLogRecords(DumpRecord)

    else:

        def events(self, svc_name):
            self.raise_error("NO_WIN32")

        def _query_service(self, svc_name):
            self.raise_error("NO_WIN32")

        def _start_service(self, config, descript, svc_name, pasw, user):
            self.raise_error("NO_WIN32")

        def _stop_service(self, svc_name):
            self.raise_error("NO_WIN32")

    #
    # End Windows Methods
    #

    def _node(self, omero_node=None):
        """
        Overrides the regular node() logic to return the value of
        OMERO_MASTER or "master"
        """
        if omero_node is not None:
            os.environ["OMERO_MASTER"] = omero_node

        if "OMERO_MASTER" in os.environ:
            return os.environ["OMERO_MASTER"]
        else:
            return "master"

    def _get_etc_dir(self):
        """Return path to directory containing configuration files"""
        return old_div(self.ctx.dir, "etc")

    def _get_grid_dir(self):
        """Return path to directory containing Gridconfiguration files"""
        return old_div(self._get_etc_dir(), "grid")

    def _get_templates_dir(self):
        """Return path to directory containing templates"""
        return self.ctx.dir / "etc" / "templates"

    def _cmd(self, *command_arguments):
        """
        Used to generate an icegridadmin command line argument list
        """
        command = ["icegridadmin", self._intcfg()]
        command.extend(command_arguments)
        return command

    def _descript(self, args):
        if args.file is not None:
            # Relative to cwd
            descript = path(args.file).abspath()
            if not descript.exists():
                self.ctx.dbg("No such file: %s -- Using as target" % descript)
                args.targets.insert(0, args.file)
                descript = None
        else:
            descript = None

        if descript is None:
            __d__ = "default.xml"
            if self._isWindows():
                __d__ = "windefault.xml"
            descript = old_div(self._get_grid_dir(), __d__)
            self.ctx.err("No descriptor given. Using %s"
                         % os.path.sep.join(["etc", "grid", __d__]))
        return descript

    def checkwindows(self, args):
        r"""
        Checks that the templates file as defined in etc\Windows.cfg
        can be found.
        """
        self.check_access(os.R_OK)
        if not self._isWindows():
            self.raise_error("NOT_WINDOWS")

        import Ice
        key = "IceGrid.Node.Data"
        properties = Ice.createProperties([self._icecfg()])
        nodedata = properties.getProperty(key)
        if not nodedata:
            self.raise_error("BAD_CONFIG")
        nodepath = path(nodedata)
        pp = nodepath.parpath(self.ctx.dir)
        if pp:
            return
        if nodepath == r"c:\omero_dist\var\master":
            self.ctx.out("Found default value: %s" % nodepath)
            self.ctx.out("Attempting to correct...")
            from omero.install.win_set_path import win_set_path
            count = win_set_path(dir=self.ctx.dir)
            if count:
                return
        self.raise_error("WIN_CONFIG", nodedata, self.ctx.dir)

    ##############################################
    #
    # Commands
    #

    @with_config
    def startasync(self, args, config):
        """
        First checks for a valid installation, then checks the grid,
        then registers the action: "node HOST start"
        """
        self.requires_service_manager(config)
        self.check_access(mask=os.R_OK, config=config)
        self.checkice()
        self.check_node(args)

        if not self.check_internal_cfg() or args.force_rewrite:
            self.rewrite(args, config, force=True)

        if 0 == self.status(args, node_only=True, can_force_rewrite=True):
            self.ctx.die(876, "Server already running")

        if not args.force_rewrite:
            self.rewrite(args, config)

        if self._isWindows():
            self.checkwindows(args)
        self.check_lock(config)

        try:
            config['omero.db.poolsize']
        except KeyError:
            self.ctx.out(
                "WARNING: Your server has not been configured for production "
                "use.\nSee https://docs.openmicroscopy.org/omero/latest/"
                "sysadmins/server-performance.html?highlight=poolsize\n"
                "for more information.")

        # Warn if deprecated TLS 1.0 and 1.1 protocols are allowed
        # See https://datatracker.ietf.org/doc/html/rfc8996
        # Both protocols are included in the default value of IceSSL.Protocols
        # https://doc.zeroc.com/ice/3.6/property-reference/icessl#id-.IceSSL.*v3.6-IceSSL.Protocols
        DEPRECATED_TLS_MESSAGE = (
            "Your server is configured to allow a deprecated TLS protocol."
            "\n\nPlease refer to https://omero.readthedocs.io/en/stable/"
            "sysadmins/server-upgrade.html#server-certificates for "
            "instructions on how to upgrade your configuration.")
        try:
            ssl_protocols = config["omero.glacier2.IceSSL.Protocols"]
            if ("TLS1_0" in ssl_protocols or "TLS1_1" in ssl_protocols):
                self.ctx.out("WARNING: " + DEPRECATED_TLS_MESSAGE)
        except KeyError:
            self.ctx.out("WARNING: " + DEPRECATED_TLS_MESSAGE)

        self._initDir()
        # Do a check to see if we've started before.
        self._regdata()
        self.check([])

        command = None
        descript = self._descript(args)
        foreground = hasattr(args, "foreground") and args.foreground

        if self._isWindows():
            if foreground:
                command = """icegridnode.exe "%s" --deploy "%s" %s\
                """ % (self._icecfg(), descript, args.targets)
            else:
                user = args.user
                pasw = args.password
                svc_name = "%s.%s" % (
                    self._get_service_name(config), args.node)
                self._start_service(config, descript, svc_name, pasw, user)
        else:
            if foreground:
                command = ["icegridnode", "--nochdir", self._icecfg(),
                           "--deploy", str(descript)] + args.targets
            else:
                command = ["icegridnode", "--daemon", "--pidfile",
                           str(self._pid()), "--nochdir", self._icecfg(),
                           "--deploy", str(descript)] + args.targets

        if command is not None:
            self.ctx.rv = self.ctx.call(command)

    @with_config
    def start(self, args, config):
        self.startasync(args, config)
        try:
            self.waitup(args)
        except NonZeroReturnCode as nzrc:
            # stop() may itself throw,
            # if it does not, then we rethrow
            # the original
            self.ctx.err('Calling "stop" on remaining components')
            self.stop(args, config)
            raise nzrc

    @with_config
    def deploy(self, args, config):
        self.rewrite(args, config)
        self.check_access()
        self.checkice()
        descript = self._descript(args)

        # TODO : Doesn't properly handle whitespace
        # Though users can workaround with something like:
        # omero admin deploy etc/grid/a\\\\ b.xml
        command = ["icegridadmin", self._intcfg(), "-e",
                   " ".join(["application", "update", str(descript)] +
                            args.targets)]
        self.ctx.call(command)

    def check_internal_cfg(self):
        internal_cfg = self._cfglist()[0]
        return os.path.exists(internal_cfg)

    @windows_warning
    def status(self, args, node_only=False, can_force_rewrite=False):
        self.check_node(args)
        if not self.check_internal_cfg():
            error_msg = 'Missing internal configuration.'
            if can_force_rewrite:
                error_msg += ' Pass --force-rewrite to the command.'
            else:
                error_msg += ' Run `omero admin rewrite`.'
            self.ctx.die(574, error_msg)
        command = self._cmd("-e", "node ping %s" % self._node())
        self.ctx.rv = self.ctx.popen(command).wait()  # popen

        # node_only implies that "up" need not check for all
        # of blitz to be accessible but just that if the node
        # is running.
        if not node_only:
            node_only = getattr(args, "nodeonly", False)

        if self.ctx.rv == 0 and not node_only:
            try:
                import Ice
                ic = Ice.initialize([self._intcfg()])
                try:
                    sm = self.session_manager(ic)
                    try:
                        sm.create("####### STATUS CHECK ########", None)
                        # Not adding "omero.client.uuid"
                    except omero.WrappedCreateSessionException:
                        # Only the server will throw one of these
                        self.ctx.dbg("Server reachable")
                        self.ctx.rv = 0
                finally:
                    ic.destroy()
            except Exception as exc:
                self.ctx.rv = 1
                self.ctx.dbg("Server not reachable: "+str(exc))

        return self.ctx.rv

    def wait_for_icedb(self, args, config):
        """
        Since the stop and start are separately protected by
        the lock on config.xml, we need to wait for some time
        to hopefully let the icegridnode process release the
        file locks.
        """
        self.ctx.sleep(1)  # put in sleep to try to prevent "db locked" (#7325)

    @with_config
    def restart(self, args, config):
        if not self.stop(args, config):
            self.ctx.die(54, "Failed to shutdown")
        self.wait_for_icedb(args, config)
        self.start(args, config)

    @with_config
    def restartasync(self, args, config):
        if not self.stop(args, config):
            self.ctx.die(54, "Failed to shutdown")
        self.wait_for_icedb(args, config)
        self.startasync(args, config)

    def waitup(self, args):
        """
        Loops 30 times with 10 second pauses waiting for status()
        to return 0. If it does not, then ctx.die() is called.
        """
        self.check_access(os.R_OK)
        self.ctx.out("Waiting on startup. Use CTRL-C to exit")
        count, loop_secs, time_msg = self.loops_and_wait(args)
        while True:
            count = count - 1
            if count == 0:
                self.ctx.die(43, "\nFailed to startup some components after"
                             " %s" % time_msg)
            elif 0 == self.status(args, node_only=False):
                break
            else:
                self.ctx.out(".", newline=False)
                self.ctx.sleep(loop_secs)

    def waitdown(self, args):
        """
        Returns true if the server went down
        """
        self.check_access(os.R_OK)
        self.ctx.out("Waiting on shutdown. Use CTRL-C to exit")
        count, loop_secs, time_msg = self.loops_and_wait(args)
        while True:
            count = count - 1
            if count == 0:
                self.ctx.die(44, "\nFailed to shutdown some components after"
                             " %s" % time_msg)
                return False
            elif 0 != self.status(args, node_only=True):
                break
            else:
                self.ctx.out(".", newline=False)
                self.ctx.sleep(loop_secs)
        self.ctx.rv = 0
        return True

    def loops_and_wait(self, args):
        """
        If present, get the wait time from the args argument
        and calculate the number of loops and the wait time
        needed. If not present in args, use a default value.
        """

        if not hasattr(args, "wait"):
            # This might happen if a new command starts using
            # waitup/waitdown without setting wait=True for
            # Action()
            args.wait = DEFAULT_WAIT

        total_secs = args.wait
        loop_secs = old_div(total_secs, 30.0)
        return 30, loop_secs, "%s seconds" % total_secs

    @with_config
    def stopasync(self, args, config):
        """
        Returns true if the server was already stopped
        """
        self.requires_service_manager(config)
        self.check_node(args)
        if args.force_rewrite:
            self.rewrite(args, config, force=True)
        if 0 != self.status(args, node_only=True, can_force_rewrite=True):
            self.ctx.err("Server not running")
            return True
        elif self._isWindows():
            svc_name = "%s.%s" % (self._get_service_name(config), args.node)
            output = self._query_service(svc_name)
            if 0 <= output.find("DOESNOTEXIST"):
                self.ctx.die(203, "%s does not exist. Use 'start' first."
                             % svc_name)
            self._stop_service(svc_name)
        else:
            command = self._cmd("-e", "node shutdown %s" % self._node())
            try:
                self.ctx.call(command)
            except NonZeroReturnCode as nzrc:
                self.ctx.rv = nzrc.rv
                self.ctx.out("Was the server already stopped?")

    @with_config
    def stop(self, args, config):
        if not self.stopasync(args, config):
            return self.waitdown(args)
        return True

    def check(self, args):
        # print "Check db. Have a way to load the db control"
        pass

    def ice(self, args):
        self.check_access()
        command = self._cmd()
        if len(args.argument) > 0:
            command.extend(["-e", " ".join(args.argument)])
            return self.ctx.call(command)
        else:
            self.ctx.call(command)

    @admin_only(full_admin=True)
    @with_config
    def fixpyramids(self, args, config):
        self.check_access()
        from omero.util.cleanse import fixpyramids
        client = self.ctx.conn(args)
        client.getSessionId()
        fixpyramids(data_dir=args.data_dir, dry_run=args.dry_run,
                    admin_service=client.sf.getAdminService(),
                    query_service=client.sf.getQueryService(),
                    config_service=client.sf.getConfigService())

    @admin_only(full_admin=True)
    def removepyramids(self, args):
        from omero.util.cleanse import removepyramids
        client = self.ctx.conn(args)
        client.getSessionId()
        if args.wait is not None and int(args.wait) > 0:
            wait = args.wait
        else:
            wait = 25
        if args.limit is not None and int(args.limit) > 0:
            limit = args.limit
        else:
            limit = 500
        if args.endian == "both":
            little = None
        elif args.endian == "little":
            little = omero.rtypes.rbool(True)
        else:
            little = omero.rtypes.rbool(False)
        # check time
        value = None
        if args.imported_after is not None:
            date = time.strptime(args.imported_after, "%Y-%m-%d")
            value = omero.rtypes.rtime(time.mktime(date)*1000)
        removepyramids(client=client,
                       little_endian=little,
                       dry_run=args.dry_run,
                       imported_after=value,
                       limit=limit, wait=wait)

    @with_config
    def jvmcfg(self, args, config):
        """Display JVM settings from the current configuration"""

        from xml.etree.ElementTree import XML
        from omero.install.jvmcfg import adjust_settings

        # JVM configuration regeneration
        templates = self._get_templates_dir() / "grid" / "templates.xml"
        template_xml = XML(templates.text())
        try:
            rv = adjust_settings(config, template_xml)
        except Exception as e:
            self.ctx.die(11, 'Cannot adjust memory settings in %s.\n%s'
                         % (templates, e))

        self.ctx.out("JVM Settings:")
        self.ctx.out("============")
        for k, v in sorted(rv.items()):
            settings = v.pop(0)
            sb = " ".join([str(x) for x in v])
            if str(settings) != "Settings()":
                sb += " # %s" % settings
            self.ctx.out("%s=%s" % (k, sb))

    def _get_omero_properties(self):
        omero_props_file = old_div(self._get_etc_dir(), "omero.properties")
        pp = PropertyParser()
        omero_props = dict(
            (p.key, p.val) for p in pp.parse_file(omero_props_file))
        if sys.platform == "darwin":
            # Override xxx.yyy properties with value of _xxx.yyy.darwin
            for key in omero_props:
                if key.startswith('_') and key.endswith('.darwin'):
                    omero_props[key[1:-7]] = omero_props[key]
        return omero_props

    def _glacier2_icessl_xml(self, config_props):
        # Convert omero.glacier2.IceSSL.* properties to IceSSL.*
        items = list(config_props.items())
        glacier2_icessl = dict((k[15:], v) for (k, v) in items
                               if k.startswith('omero.glacier2.IceSSL.'))
        return ['<property name="%s" value="%s"/>' % kv
                for kv in list(glacier2_icessl.items())]

    def get_omero_server_version(self):
        """Returns the value of omero.version stored in omero.properties"""
        omero_props_file = os.path.join(self._get_etc_dir(), "omero.properties")
        with open(omero_props_file, 'r') as f:
            for line in f.readlines():
                if line.startswith("omero.version"):
                    idx = line.index("=")
                    return line[idx + 1:].rstrip()
        return "unknown"

    @with_config
    def rewrite(self, args, config, force=False):
        """
        Regenerate all the template files under the configuration directory
        """

        from xml.etree.ElementTree import XML
        from omero.install.jvmcfg import adjust_settings

        if self.check_internal_cfg() and not force:
            if 0 == self.status(args, node_only=True):
                self.ctx.die(
                    100, "Can't regenerate templates the server is running!")
            # Reset return value
            self.ctx.rv = 0

        # JVM configuration regeneration
        templates = self._get_templates_dir()/"grid"/"templates.xml"

        # Get some defaults from omero.properties
        config_props = self._get_omero_properties()

        generated = old_div(self._get_grid_dir(), "templates.xml")
        if generated.exists():
            generated.remove()
        config2 = omero.config.ConfigXml(str(generated))

        config_props.update(config.as_map())
        template_xml_text = templates.text().replace(
            '@omero.glacier2.icessl@',
            '\n'.join(self._glacier2_icessl_xml(config_props)))
        template_xml = XML(template_xml_text)

        try:
            rv = adjust_settings(config, template_xml)
        except Exception as e:
            self.ctx.die(11, 'Cannot adjust memory settings in %s.\n%s'
                         % (templates, e))

        def clear_tail(elem):
            elem.tail = ""
            if elem.text is not None and not elem.text.strip():
                elem.text = ""
            for child in list(elem):
                clear_tail(child)

        clear_tail(template_xml)
        config2.write_element(template_xml)
        config2.XML = None  # Prevent re-saving
        config2.close()
        config.save()

        # Define substitution dictionary for template files
        config = config.as_map()
        substitutions = {
            '@omero.ports.prefix@': config.get('omero.ports.prefix', ''),
            '@omero.ports.ssl@': config.get('omero.ports.ssl', '4064'),
            '@omero.ports.tcp@': config.get('omero.ports.tcp', '4063'),
            '@omero.ports.wss@': config.get('omero.ports.wss', '4066'),
            '@omero.ports.ws@': config.get('omero.ports.ws', '4065'),
            '@omero.ports.registry@': config.get(
                'omero.ports.registry', '4061'),
            '@omero.master.host@': config.get('omero.master.host', config.get(
                'Ice.Default.Host', '127.0.0.1'))
            }

        client_transports = config.get('omero.client.icetransports', 'ssl,tcp')
        client_endpoints = [
            '{tp} -p {prefix}{port}'.format(
                tp=clienttp,
                prefix=substitutions['@omero.ports.prefix@'],
                port=substitutions['@omero.ports.{tp}@'.format(tp=clienttp)],
            )
            for clienttp in client_transports.split(',')]
        substitutions['@omero.client.endpoints@'] = ':'.join(client_endpoints)

        node_descriptors = config.get('omero.server.nodedescriptors', '')

        def copy_template(input_file, output_dir, post_process=None):
            """Replace templates"""
            with open(input_file) as template:
                data = template.read()
            output_file = path(old_div(output_dir,
                               os.path.basename(input_file)))
            if output_file.exists():
                output_file.remove()
            with open(output_file, 'w') as f:
                for key, value in substitutions.items():
                    data = re.sub(key, value, data)
                if post_process:
                    data = post_process(data)
                f.write(data)

        # Regenerate various configuration files from templates
        for cfg_file in glob(old_div(self._get_templates_dir(), "*.cfg")):
            copy_template(cfg_file, self._get_etc_dir())
        for xml_file in glob(
                self._get_templates_dir() / "grid" / "*default.xml"):
            copy_template(
                xml_file, old_div(self._get_etc_dir(), "grid"),
                lambda xml: _process_xml(xml, node_descriptors))
        ice_config = old_div(self._get_templates_dir(), "ice.config")
        substitutions['@omero.master.host@'] = config.get(
            'omero.master.host', config.get('Ice.Default.Host', 'localhost'))
        copy_template(ice_config, self._get_etc_dir())

        return rv

    @windows_warning
    @with_config
    def diagnostics(self, args, config):

        self._diagnostics_banner("admin")

        from xml.etree.ElementTree import XML
        from omero.install.jvmcfg import read_settings

        self.check_access(os.R_OK)
        templates = old_div(self._get_grid_dir(), "templates.xml")
        if templates.exists():
            template_xml = XML(templates.text())
            try:
                memory = read_settings(template_xml)
            except Exception as e:
                self.ctx.die(11, 'Cannot read memory settings in %s.\n%s'
                             % (templates, e))
        else:
            memory = None
        omero_data_dir = self._get_data_dir(config)

        from omero.util.temp_files import gettempdir
        # gettempdir returns ~/omero/tmp/omero_%NAME/%PROCESS
        # To find something more generally useful for calculating
        # size, we go up two directories
        omero_temp_dir = gettempdir()
        omero_temp_dir = os.path.abspath(
            os.path.join(omero_temp_dir, os.path.pardir, os.path.pardir))

        def version(cmd):
            """
            Returns a true response only
            if a valid version was found.
            """
            self._item("Commands", "%s" % " ".join(cmd))
            try:
                p = self.ctx.popen(cmd)
            except OSError:
                self.ctx.err("not found")
                return False

            p.wait()
            io = list(map(bytes_to_native_str, p.communicate()))
            try:
                v = io[0].split()
                v.extend(io[1].split())
                v = "".join(v)
                m = re.match(r"^\D*(\d[.\d]+\d)\D?.*$", v)
                v = "%-10s" % m.group(1)
                self.ctx.out(v, False)
                try:
                    where = whichall(cmd[0])
                    sz = len(where)
                    if sz == 0:
                        where = "unknown"
                    else:
                        where = where[0]
                        if sz > 1:
                            where += " -- %s others" % sz
                except Exception:
                    where = "unknown"
                self.ctx.out("(%s)" % where)
                return True
            except Exception as e:
                self.ctx.err("error:%s" % e)
                return False

        import logging
        logging.basicConfig()
        check = self.run_upgrade_check(config, "diagnostics")
        if check.isUpgradeNeeded():
            self.ctx.out("")

        version(["java",         "-version"])
        version(["python",       "-V"])
        version(["icegridnode",  "--version"])
        iga = version(["icegridadmin", "--version"])
        version(["psql",         "--version"])
        version(["openssl",      "version"])

        self.ctx.out("")
        self._item("Component", "OMERO.py")
        self.ctx.out(VERSION)
        self._item("Component", "OMERO.server")
        self.ctx.out(self.get_omero_server_version())

        def get_ports(input):
            router_lines = [line for line in input.split("\n")
                            if line.find("ROUTER") >= 0]

            ssl_port = None
            tcp_port = None
            for line in router_lines:
                if not ssl_port and line.find("ROUTERPORT") >= 0:
                    m = re.match(r".*?(\d+).*?$", line)
                    if m:
                        ssl_port = m.group(1)

                if not tcp_port and line.find("INSECUREROUTER") >= 0:
                    m = re.match(r"^.*?-p (\d+).*?$", line)
                    if m:
                        tcp_port = m.group(1)
            return ssl_port, tcp_port

        self.ctx.out("")
        if not iga:
            self.ctx.out(
                "No icegridadmin available: Cannot check server list")
        else:
            self._item("Server", "icegridnode")
            p = self.ctx.popen(self._cmd("-e", "server list"))  # popen
            rv = p.wait()
            io = p.communicate()
            if rv != 0:
                self.ctx.out("not started")
                self.ctx.dbg("""
                Stdout:\n%s
                Stderr:\n%s
                """ % io)
            else:
                self.ctx.out("running")
                servers = io[0].split()
                servers.sort()
                for s in servers:
                    decoded = s.decode("utf-8")
                    self._item("Server", "%s" % decoded)
                    p2 = self.ctx.popen(
                        self._cmd("-e", "server state %s" % decoded))  # popen
                    p2.wait()
                    io2 = p2.communicate()
                    if io2[1]:
                        self.ctx.err(io2[1].strip().decode("utf-8"))
                    elif io2[0]:
                        self.ctx.out(io2[0].strip().decode("utf-8"))
                    else:
                        self.ctx.err("UNKNOWN!")
            if self._isWindows():
                # Print the OMERO server Windows service details
                hscm = win32service.OpenSCManager(
                    None, None, win32service.SC_MANAGER_ALL_ACCESS)
                services = win32service.EnumServicesStatus(hscm)
                omesvcs = tuple((sname, fname) for sname, fname, status
                                in services if "OMERO" in fname)
                for sname, fname in omesvcs:
                    self._item("Server", fname)
                    hsc = win32service.OpenService(
                        hscm, sname, win32service.SC_MANAGER_ALL_ACCESS)
                    logonuser = win32service.QueryServiceConfig(hsc)[7]
                    if win32service.QueryServiceStatus(hsc)[1] == \
                            win32service.SERVICE_RUNNING:
                        self.ctx.out("active (running as %s)" % logonuser)
                    else:
                        self.ctx.out("inactive")
                    win32service.CloseServiceHandle(hsc)
                win32service.CloseServiceHandle(hscm)

        def parse_logs():

            log_dir = self.ctx.dir / "var" / "log"
            self.ctx.out("")
            self._item("Log dir", "%s" % log_dir.abspath())
            if not log_dir.exists():
                self.ctx.out("")
                self.ctx.out("No logs available")
                return
            else:
                self._exists(log_dir)

            known_log_files = [
                "Blitz-0.log", "Tables-0.log", "Processor-0.log",
                "Indexer-0.log", "FileServer.log", "MonitorServer.log",
                "DropBox.log", "TestDropBox.log"]
            files = log_dir.files()
            files = set([x.basename() for x in files])
            # Adding known names just in case
            for x in known_log_files:
                files.add(x)
            files = list(files)
            files.sort()
            for x in files:
                self._item("Log files", x)
                self._exists(old_div(log_dir, x))
            self._item("Log files", "Total size")
            sz = 0
            for x in log_dir.walkfiles():
                sz += x.size
            self.ctx.out("%-.2f MB" % (old_div(float(sz), 1000000.0)))
            self.ctx.out("")

            # Parsing well known issues
            ready = re.compile(r".*?ome.services.util.ServerVersionCheck\
            .*OMERO.Version.*Ready..*?")
            db_ready = re.compile(r".*?Did.you.create.your.database[?].*?")
            data_dir = re.compile(r".*?Unable.to.initialize:.FullText.*?")
            pg_password = re.compile(r".*?org.postgresql.util.PSQLException:\
            .FATAL:.password.*?authentication.failed.for.user.*?")
            pg_user = re.compile(r""".*?org.postgresql.util.PSQLException:\
            .FATAL:.role.".*?".does.not.exist.*?""")
            pg_conn = re.compile(r""".*?org.postgresql.util.PSQLException:\
            .Connection.refused.""")

            issues = {
                ready: "=> Server restarted <=",
                db_ready: "Your database configuration is invalid",
                data_dir: "Did you create your omero.data.dir? E.g. /OMERO",
                pg_password: "Your postgres password seems to be invalid",
                pg_user: "Your postgres user is invalid",
                pg_conn: "Your postgres hostname and/or port is invalid"
            }

            try:
                for file in ('Blitz-0.log',):

                    p = self.ctx.dir / "var" / "log" / file
                    import fileinput
                    for line in fileinput.input([str(p)]):
                        lno = fileinput.filelineno()
                        for k, v in list(issues.items()):
                            if k.match(line):
                                self._item('Parsing %s' % file,
                                           "[line:%s] %s" % (lno, v))
                                self.ctx.out("")
                                break
            except Exception:
                self.ctx.err("Error while parsing logs")

        if not args.no_logs:
            parse_logs()

        self.ctx.out("")

        def env_val(val):
            self._item("Environment", "%s=%s"
                       % (val, os.environ.get(val, "(unset)")))
            self.ctx.out("")
        env_val("OMERO_HOME")
        env_val("OMERODIR")
        env_val("OMERO_NODE")
        env_val("OMERO_MASTER")
        env_val("OMERO_USERDIR")
        env_val("OMERO_TMPDIR")
        env_val("PATH")
        env_val("PYTHONPATH")
        env_val("ICE_HOME")
        env_val("LD_LIBRARY_PATH")
        env_val("DYLD_LIBRARY_PATH")

        # List SSL & TCP ports of deployed applications
        self.ctx.out("")
        p = self.ctx.popen(self._cmd("-e", "application list"))  # popen
        rv = p.wait()
        io = p.communicate()
        if rv != 0:
            self.ctx.out("Cannot list deployed applications.")
            self.ctx.dbg("""
            Stdout:\n%s
            Stderr:\n%s
            """ % io)
        else:
            applications = io[0].split()
            applications.sort()
            for s in applications:
                def port_val(port_type, value):
                    self._item("%s %s port" % (s, port_type),
                               "%s" % value or "Not found")
                    self.ctx.out("")
                s = s.decode("utf-8")
                p2 = self.ctx.popen(
                    self._cmd("-e", "application describe %s" % s))
                io2 = p2.communicate()
                if io2[1]:
                    self.ctx.err(io2[1].strip().decode("utf-8"))
                elif io2[0]:
                    ssl_port, tcp_port = get_ports(io2[0].decode("utf-8"))
                    port_val("SSL", ssl_port)
                    port_val("TCP", tcp_port)
                else:
                    self.ctx.err("UNKNOWN!")

        for dir_name, dir_path, dir_size in (
                ("data", omero_data_dir, ""),
                ("temp", omero_temp_dir, True)):
            dir_path_exists = os.path.exists(dir_path)
            is_writable = os.access(dir_path, os.R_OK | os.W_OK)
            if dir_size and dir_path_exists:
                dir_size = self.getdirsize(omero_temp_dir, False)
                dir_size = "   (Size: %s)" % dir_size
            self._item("OMERO %s dir" % dir_name, "'%s'" % dir_path)
            self.ctx.out("Exists? %s\tIs writable? %s%s" %
                         (dir_path_exists, is_writable,
                          dir_size))

        # JVM settings
        self.ctx.out("")
        if memory:
            for k, v in sorted(memory.items()):
                sb = " ".join([str(x) for x in v])
                self._item("JVM settings", " %s" % (k[0].upper() + k[1:]))
                self.ctx.out("%s" % sb)

        def jar_manifest(jar):
            manifest = {}
            error = ''
            try:
                with ZipFile(jar) as z:
                    current = ''
                    for line in z.read('META-INF/MANIFEST.MF').splitlines():
                        line = line.decode()
                        if line and line[0] == ' ':
                            current += line[1:]
                        else:
                            if current:
                                manifest.update([current.split(': ', 1)])
                            current = line
                    if current:
                        manifest.update([current.split(': ', 1)])
            except Exception as e:
                error = str(e)
            return manifest, error

        # Jar versions
        jars = sorted(
            os.path.join(os.path.relpath(root, self.ctx.dir), filename)
            for root, dirnames, filenames in os.walk(self.ctx.dir)
            for filename in filenames
            if filename.endswith('.jar')
        )
        if args.all_jars:
            jar_re = r'.*'
        else:
            jar_re = r'lib/server/(formats|ome|omero)-.*.jar'
        manifest_keys = (
            'Implementation-Title',
            'Implementation-Version',
            'Implementation-Date',
            'Implementation-Build',
        )
        self.ctx.out("")
        for jar in jars:
            if not re.match(jar_re, jar):
                continue
            manifest, error = jar_manifest(self.ctx.dir / jar)
            if error:
                info = [error]
            else:
                info = [manifest.get(key, '') for key in manifest_keys]
            self._item("Jar", jar)
            self.ctx.out('\t'.join(info))



    def email(self, args):
        client = self.ctx.conn(args)
        iadmin = client.sf.getAdminService()
        users, groups = self.get_users_groups(args, iadmin)

        if not args.text:
            args.text = ("-")

        text = " ".join(args.text)
        if text == "-":
            stdin = FileType("r")("-")
            text = stdin.read()

        if args.everyone:
            if users or groups:
                self.ctx.err("Warning: users and groups ignored")

        req = omero.cmd.SendEmailRequest(
            subject=args.subject,
            body=text,
            userIds=users,
            groupIds=groups,
            everyone=args.everyone,
            inactive=args.inactive)

        ms = 500
        wait = args.wait if args.wait > 0 else 25
        loops = ceil(wait * 1000.0 / ms)

        try:
            cb = client.submit(
                req, loops=loops, ms=ms,
                failonerror=True, failontimeout=True)
        except omero.CmdError as ce:
            err = ce.err
            if err.name == "no-body" and err.parameters:
                sb = list(err.parameters.items())
                sb = ["%s:%s" % (k, v) for k, v in sb]
                sb = "\n".join(sb)
                self.ctx.die(12, sb)
            self.ctx.die(13, "Failed to send emails:\n%s" % err)

        try:
            rsp = cb.getResponse()
            self.ctx.out(
                "Successfully sent %s of %s emails" % (
                    rsp.success, rsp.total
                ))
            if rsp.invalidusers:
                self.ctx.out(
                    "%s users had no email address" % len(
                        rsp.invalidusers)
                )
            if rsp.invalidemails:
                self.ctx.out(
                    "%s email addresses were invalid" % len(
                        rsp.invalidemails)
                )
        finally:
            cb.close(True)

    def getdirsize(self, directory, strict=True):
        """
        Uses os.walk to calculate the deep size of the given
        directory. If strict is set, an exception will be thrown
        if a directory disappears during calculation. Otherwise,
        an error will be printed to stderr, but the calculation
        will continue.
        """
        total = 0
        for values in os.walk(directory):
            for filename in values[2]:
                fullpath = os.path.join(values[0], filename)
                try:
                    total += os.path.getsize(fullpath)
                except Exception:
                    self.ctx.err("Failed to get size of: %s" % fullpath)
                    if strict:
                        raise
        return total

    def session_manager(self, communicator):
        import IceGrid
        import Glacier2
        iq = communicator.stringToProxy("IceGrid/Query")
        iq = IceGrid.QueryPrx.checkedCast(iq)
        sm = iq.findAllObjectsByType("::Glacier2::SessionManager")[0]
        sm = Glacier2.SessionManagerPrx.checkedCast(sm)
        return sm

    def can_access(self, filepath, mask=os.R_OK | os.W_OK):
        """
        Check that the given path belongs to
        or is accessible by the current user
        on Linux systems.
        """

        if "Windows" == platform.system():
            return

        pathobj = path(filepath)

        if not pathobj.exists():
            self.ctx.die(8, "FATAL: OMERO directory does not exist: %s"
                         % pathobj)

        if not os.access(filepath, mask):
            self.ctx.die(10, "FATAL: Cannot access %s, a required"
                         " file/directory for OMERO" % filepath)

    def check_access(self, mask=os.R_OK | os.W_OK, config=None):
        """Check that 'var' is accessible by the current user."""

        var = old_div(self.ctx.dir, 'var')
        if not os.path.exists(var):
            self.ctx.out("Creating directory %s" % var)
            os.makedirs(var)
        else:
            self.can_access(var, mask)

        if config is not None:
            omero_data_dir = self._get_data_dir(config)
            self.can_access(omero_data_dir, mask)

        for p in os.listdir(var):
            subpath = os.path.join(var, p)
            if os.path.isdir(subpath):
                self.can_access(subpath, mask)

    def check_lock(self, config):
        """
        Issue a warning if any of the top ".omero" directories
        contain a lock file. This isn't a conclusive test this
        we don't have access to the DB to get the UUID
        for this instance. Usually there should only be one
        though.
        """
        omero_data_dir = self._get_data_dir(config)
        lock_files = os.path.join(
            omero_data_dir, ".omero", "repository",
            "*", ".lock")
        lock_files = glob(lock_files)
        if lock_files:
            self.ctx.err("WARNING: lock files in %s" %
                         omero_data_dir)
            self.ctx.err("-"*40)
            for lock_file in lock_files:
                self.ctx.err(lock_file)
            self.ctx.err("-"*40)
            self.ctx.err((
                "\n"
                "You may want to stop all server processes and remove\n"
                "these files manually. Lock files can remain after an\n"
                "abrupt server outage and are especially frequent on\n"
                "remotely mounted filesystems like NFS.\n"))

    def check_node(self, args):
        """
        If the args argparse.Namespace argument has no "node" attribute,
        then assign one.
        """
        if not hasattr(args, "node"):
            args.node = self._node()

    def checkice(self, args=None):
        """
        Checks for Ice version compatibility.

        See ticket:2514, ticket:1260
        """

        def _check(msg, vers):
            compat = ice_compatibility.split(".")
            if isbytes(vers):
                vers = bytes_to_native_str(vers)
            vers = vers.split(".")
            if compat[0:2] != vers[0:2]:
                self.ctx.die(164, "%s is not compatible with %s: %s"
                             % (msg, ".".join(compat), ".".join(vers)))

        import Ice
        vers = Ice.stringVersion()
        _check("IcePy version", vers)

        # See ticket #10051
        popen = self.ctx.popen(["icegridnode", "--version"])
        env = self.ctx._env()
        # Unclear how this could have been set with the call to unsetenv
        ice_config = env.get("ICE_CONFIG")
        if ice_config is not None and not os.path.exists(ice_config):
            popen = self.ctx.popen(["icegridnode", "--version"],
                                   **{'ICE_CONFIG': ''})

        vers = popen.communicate()[1]
        _check("icegridnode version", vers)

    def open_config(self, unused=None):
        """
        Callers are responsible for closing the
        returned ConfigXml object.
        """
        cfg_xml = old_div(self._get_grid_dir(), "config.xml")
        cfg_tmp = old_div(self._get_grid_dir(), "config.xml.tmp")
        grid_dir = self._get_grid_dir()
        if not cfg_xml.exists() and self.can_access(grid_dir):
            if cfg_tmp.exists() and self.can_access(cfg_tmp):
                self.ctx.dbg("Removing old config.xml.tmp")
                cfg_tmp.remove()
            config = omero.config.ConfigXml(str(cfg_tmp))
            try:
                self.ctx.controls["config"].upgrade(None, config)
            finally:
                config.close()
            self.ctx.err("Creating %s" % cfg_xml)
            cfg_tmp.rename(str(cfg_xml))

        try:
            try:
                config = omero.config.ConfigXml(str(cfg_xml))
            except Exception as e:
                self.ctx.die(577, str(e))
            if config.save_on_close:
                config.save()
            else:
                self.ctx.err("%s read-only" % cfg_xml)
        except portalocker.LockException:
            try:
                config.close()
            except Exception:
                pass
            self.ctx.die(111, "Could not acquire lock on %s" % cfg_xml)

        return config

    @with_config
    def reindex(self, args, config):

        self.check_access(config=config)
        import omero.java
        server_dir = self.ctx.dir / "lib" / "server"
        log_config_file = self.ctx.dir / "etc" / "logback-indexing-cli.xml"
        logback = "-Dlogback.configurationFile=%s" % log_config_file
        classpath = [file.abspath() for file in server_dir.files("*.jar")]
        xargs = [logback, "-cp", os.pathsep.join(classpath)]
        # See etc/grid/templates.xml
        for v in (("warn", "3600000"), ("error", "86400000")):
            xargs.append("-Domero.throttling.method_time.%s=%s" % v)

        cfg = config.as_map()
        omero_data_dir = self._get_data_dir(config)
        config.close()  # Early close. See #9800
        for x in ("name", "user", "host", "port", "properties", "url"):
            # NOT passing password on command-line
            k = "omero.db.%s" % x
            if k in cfg:
                v = cfg[k]
                xargs.append("-D%s=%s" % (k, v))

        xargs.append("-Domero.data.dir=%s" % omero_data_dir)
        for k, v in list(cfg.items()):
            if k.startswith("omero.search"):
                xargs.append("-D%s=%s" % (k, cfg[k]))

        locks = {"native": "org.apache.lucene.store.NativeFSLockFactory",
                 "simple": "org.apache.lucene.store.SimpleFSLockFactory",
                 "none": "org.apache.lucene.store.NoLockFactory"}
        lock = locks.get(args.lock_factory, args.lock_factory)

        year2 = datetime.datetime.now().year + 2
        factory_class = "org.apache.lucene.store.FSDirectoryLockFactoryClass"
        xargs2 = ["-Xmx%s" % args.mem,
                  "-Domero.search.cron=1 1 1 1 1 ? %s" % year2,
                  "-Domero.search.batch=%s" % args.batch,
                  "-Domero.search.merge_factor=%s" % args.merge_factor,
                  "-Domero.search.ram_buffer_size=%s" % args.ram_buffer_size,
                  "-D%s=%s" % (factory_class, lock)]
        xargs.extend(xargs2)

        cmd = ["ome.services.fulltext.Main"]

        # Python actions
        early_exit = False
        if args.wipe:
            early_exit = True
            self.can_access(omero_data_dir)
            from os.path import sep
            pattern = sep.join([omero_data_dir, "FullText", "*"])
            files = glob(pattern)
            total = 0
            self.ctx.err("Wiping %s files matching %s" % (len(files), pattern))
            for file in files:
                size = os.path.getsize(file)
                total += 0
                print(file, size)
            print("Total:", size)
            yes = self.ctx.input("Enter 'y' to continue:")
            if not yes.lower().startswith("y"):
                return
            else:
                for file in files:
                    try:
                        os.remove(file)
                    except Exception:
                        self.ctx.err("Failed to remove: %s", file)

        elif args.prepare:
            early_exit = True
            self.stop_service("Indexer-0")
        elif args.finish:
            early_exit = True
            self.start_service("Indexer-0")

        if early_exit:
            return  # Early exit!

        if self.check_service("Indexer-0") and not args.prepare:
            self.ctx.die(578, "Indexer-0 is running")

        # Java actions
        if args.full:
            cmd.append("full")
        elif args.dryrun:
            cmd.append("dryrun")
        elif args.foreground:
            cmd.append("foreground")
            self.ctx.die(877, "foreground indexing is not available")
        elif args.reset is not None:
            cmd.append("reset")
            cmd.append(args.reset)
        elif args.events:
            cmd.append("events")
        elif getattr(args, "class"):
            cmd.append("reindex")
            cmd.extend(getattr(args, "class"))
        else:
            self.ctx.die(502, "No valid action: %s" % args)

        debug = False
        if getattr(args, "jdwp"):
            debug = True

        # Pass omero.db.pass using JAVA_OPTS environment variable
        if "omero.db.pass" in cfg:
            dbpassargs = "-Domero.db.pass=%s" % cfg["omero.db.pass"]
            if "JAVA_OPTS" not in os.environ:
                os.environ['JAVA_OPTS'] = dbpassargs
            else:
                os.environ['JAVA_OPTS'] = "%s %s" % (
                    os.environ.get('JAVA_OPTS'), dbpassargs)

        self.ctx.dbg(
            "Launching Java: %s, debug=%s, xargs=%s" % (cmd, debug, xargs))
        p = omero.java.run(cmd,
                           use_exec=True, debug=debug, xargs=xargs,
                           stdout=sys.stdout, stderr=sys.stderr)
        self.ctx.rv = p.wait()

    @admin_only(full_admin=True)
    def cleanse(self, args):
        self.check_access()
        from omero.util.cleanse import cleanse
        cleanse(data_dir=args.data_dir, client=self.ctx.conn(args),
                dry_run=args.dry_run)

    @admin_only(full_admin=False)
    def log(self, args):
        client = self.ctx.conn(args)
        req = RawAccessRequest(command='log', path=args.level,
                               repoUuid=args.repo, args=[args.message])
        client.submit(req).loop(100, 100)

    def sessionlist(self, args):
        self.ctx.err('WARNING: "admin sessionlist" is deprecated, '
                     'use "sessions who" instead')
        client = self.ctx.conn(args)
        service = client.sf.getQueryService()
        params = omero.sys.ParametersI()
        query = "select s from Session s join fetch s.node n join fetch"\
            " s.owner o where s.closed is null and n.id != 0"
        results = service.findAllByQuery(query, params)
        mapped = list()
        for s in results:
            rv = list()
            mapped.append(rv)
            if not s.isLoaded():
                rv.append("")
                rv.append("id=%s" % s.id.val)
                rv.append("")
                rv.append("")
                rv.append("")
                rv.append("insufficient privileges")
            else:
                rv.append(s.node.id)
                rv.append(s.uuid)
                rv.append(s.started)
                rv.append(s.owner.omeName)
                if s.userAgent is None:
                    rv.append("")
                else:
                    rv.append(s.userAgent)
                if client.getSessionId() == s.uuid.val:
                    rv.append("current session")
                else:
                    rv.append("")
        self.ctx.controls["hql"].display(
            mapped, ("node", "session", "started", "owner", "agent", "notes"))

    def check_service(self, name):
        command = self._cmd()
        command.extend(["-e", "server pid %s" % name])
        p = self.ctx.popen(command)  # popen
        rc = p.wait()
        return rc == 0

    def start_service(self, name):
        command = self._cmd()
        command.extend(["-e", "server enable %s" % name])
        rc = self.ctx.call(command)
        if rc != 0:
            self.ctx.err("%s could not be enabled" % name)
        else:
            self.ctx.err("%s restarted" % name)

    def stop_service(self, name):
        command = self._cmd()
        command.extend(["-e", "server disable %s" % name])
        rc = self.ctx.call(command)
        if rc != 0:
            self.ctx.err("%s may already be disabled" % name)
        else:
            command = self._cmd()
            command.extend(["-e", "server stop %s" % name])
            rc = self.ctx.call(command)
            if rc != 0:
                self.ctx.err("'server stop %s' failed" % name)
            else:
                self.ctx.err("%s stopped" % name)

    def _get_data_dir(self, config):
        config = config.as_map()
        return config.get("omero.data.dir", "/OMERO")

    @with_config
    def checkupgrade(self, args, config):
        """
        Checks whether a server upgrade is available, exits with return code
        0: this is the latest version
        1: an upgrade is available
        2: an error occurred whilst checking
        """

        uc = self.run_upgrade_check(config, CHECKUPGRADE_USERAGENT)
        if uc.isUpgradeNeeded():
            self.ctx.die(1, uc.getUpgradeUrl())
        if uc.isExceptionThrown():
            self.ctx.die(2, uc.getExceptionThrown())


try:
    register("admin", AdminControl, HELP)
except NameError:
    if __name__ == "__main__":
        cli = CLI()
        cli.register("admin", AdminControl, HELP)
        cli.invoke(sys.argv[1:])



# -*- coding: utf-8 -*-
# **********************************************************************
#
# Copyright (c) 2003-2017 ZeroC, Inc. All rights reserved.
#
# This copy of Ice is licensed to you under the terms described in the
# ICE_LICENSE file included in this distribution.
#
# **********************************************************************
#
# Ice version 3.6.4
#
# <auto-generated>
#
# Generated from file `ContrastStretchingContext.ice'
#
# Warning: do not edit this file.
#
# </auto-generated>
#

from sys import version_info as _version_info_
import Ice, IcePy
import omero_model_IObject_ice
import omero_RTypes_ice
import omero_model_RTypes_ice
import omero_System_ice
import omero_Collections_ice
import omero_model_CodomainMapContext_ice

# Included module omero
_M_omero = Ice.openModule('omero')

# Included module omero.model
_M_omero.model = Ice.openModule('omero.model')

# Included module Ice
_M_Ice = Ice.openModule('Ice')

# Included module omero.sys
_M_omero.sys = Ice.openModule('omero.sys')

# Included module omero.api
_M_omero.api = Ice.openModule('omero.api')

# Start of module omero
__name__ = 'omero'

# Start of module omero.model
__name__ = 'omero.model'

if 'ChannelBinding' not in _M_omero.model.__dict__:
    _M_omero.model._t_ChannelBinding = IcePy.declareClass('::omero::model::ChannelBinding')
    _M_omero.model._t_ChannelBindingPrx = IcePy.declareProxy('::omero::model::ChannelBinding')

if 'Details' not in _M_omero.model.__dict__:
    _M_omero.model._t_Details = IcePy.declareClass('::omero::model::Details')
    _M_omero.model._t_DetailsPrx = IcePy.declareProxy('::omero::model::Details')

if 'ContrastStretchingContext' not in _M_omero.model.__dict__:
    _M_omero.model.ContrastStretchingContext = Ice.createTempClass()
    class ContrastStretchingContext(_M_omero.model.CodomainMapContext):
        def __init__(self, _id=None, _details=None, _loaded=False, _version=None, _channelBinding=None, _xstart=None, _ystart=None, _xend=None, _yend=None):
            if Ice.getType(self) == _M_omero.model.ContrastStretchingContext:
                raise RuntimeError('omero.model.ContrastStretchingContext is an abstract class')
            _M_omero.model.CodomainMapContext.__init__(self, _id, _details, _loaded, _version, _channelBinding)
            self._xstart = _xstart
            self._ystart = _ystart
            self._xend = _xend
            self._yend = _yend

        def ice_ids(self, current=None):
            return ('::Ice::Object', '::omero::model::CodomainMapContext', '::omero::model::ContrastStretchingContext', '::omero::model::IObject')

        def ice_id(self, current=None):
            return '::omero::model::ContrastStretchingContext'

        def ice_staticId():
            return '::omero::model::ContrastStretchingContext'
        ice_staticId = staticmethod(ice_staticId)

        def getXstart(self, current=None):
            pass

        def setXstart(self, theXstart, current=None):
            pass

        def getYstart(self, current=None):
            pass

        def setYstart(self, theYstart, current=None):
            pass

        def getXend(self, current=None):
            pass

        def setXend(self, theXend, current=None):
            pass

        def getYend(self, current=None):
            pass

        def setYend(self, theYend, current=None):
            pass

        def __str__(self):
            return IcePy.stringify(self, _M_omero.model._t_ContrastStretchingContext)

        __repr__ = __str__

    _M_omero.model.ContrastStretchingContextPrx = Ice.createTempClass()
    class ContrastStretchingContextPrx(_M_omero.model.CodomainMapContextPrx):

        def getXstart(self, _ctx=None):
            return _M_omero.model.ContrastStretchingContext._op_getXstart.invoke(self, ((), _ctx))

        def begin_getXstart(self, _response=None, _ex=None, _sent=None, _ctx=None):
            return _M_omero.model.ContrastStretchingContext._op_getXstart.begin(self, ((), _response, _ex, _sent, _ctx))

        def end_getXstart(self, _r):
            return _M_omero.model.ContrastStretchingContext._op_getXstart.end(self, _r)

        def setXstart(self, theXstart, _ctx=None):
            return _M_omero.model.ContrastStretchingContext._op_setXstart.invoke(self, ((theXstart, ), _ctx))

        def begin_setXstart(self, theXstart, _response=None, _ex=None, _sent=None, _ctx=None):
            return _M_omero.model.ContrastStretchingContext._op_setXstart.begin(self, ((theXstart, ), _response, _ex, _sent, _ctx))

        def end_setXstart(self, _r):
            return _M_omero.model.ContrastStretchingContext._op_setXstart.end(self, _r)

        def getYstart(self, _ctx=None):
            return _M_omero.model.ContrastStretchingContext._op_getYstart.invoke(self, ((), _ctx))

        def begin_getYstart(self, _response=None, _ex=None, _sent=None, _ctx=None):
            return _M_omero.model.ContrastStretchingContext._op_getYstart.begin(self, ((), _response, _ex, _sent, _ctx))

        def end_getYstart(self, _r):
            return _M_omero.model.ContrastStretchingContext._op_getYstart.end(self, _r)

        def setYstart(self, theYstart, _ctx=None):
            return _M_omero.model.ContrastStretchingContext._op_setYstart.invoke(self, ((theYstart, ), _ctx))

        def begin_setYstart(self, theYstart, _response=None, _ex=None, _sent=None, _ctx=None):
            return _M_omero.model.ContrastStretchingContext._op_setYstart.begin(self, ((theYstart, ), _response, _ex, _sent, _ctx))

        def end_setYstart(self, _r):
            return _M_omero.model.ContrastStretchingContext._op_setYstart.end(self, _r)

        def getXend(self, _ctx=None):
            return _M_omero.model.ContrastStretchingContext._op_getXend.invoke(self, ((), _ctx))

        def begin_getXend(self, _response=None, _ex=None, _sent=None, _ctx=None):
            return _M_omero.model.ContrastStretchingContext._op_getXend.begin(self, ((), _response, _ex, _sent, _ctx))

        def end_getXend(self, _r):
            return _M_omero.model.ContrastStretchingContext._op_getXend.end(self, _r)

        def setXend(self, theXend, _ctx=None):
            return _M_omero.model.ContrastStretchingContext._op_setXend.invoke(self, ((theXend, ), _ctx))

        def begin_setXend(self, theXend, _response=None, _ex=None, _sent=None, _ctx=None):
            return _M_omero.model.ContrastStretchingContext._op_setXend.begin(self, ((theXend, ), _response, _ex, _sent, _ctx))

        def end_setXend(self, _r):
            return _M_omero.model.ContrastStretchingContext._op_setXend.end(self, _r)

        def getYend(self, _ctx=None):
            return _M_omero.model.ContrastStretchingContext._op_getYend.invoke(self, ((), _ctx))

        def begin_getYend(self, _response=None, _ex=None, _sent=None, _ctx=None):
            return _M_omero.model.ContrastStretchingContext._op_getYend.begin(self, ((), _response, _ex, _sent, _ctx))

        def end_getYend(self, _r):
            return _M_omero.model.ContrastStretchingContext._op_getYend.end(self, _r)

        def setYend(self, theYend, _ctx=None):
            return _M_omero.model.ContrastStretchingContext._op_setYend.invoke(self, ((theYend, ), _ctx))

        def begin_setYend(self, theYend, _response=None, _ex=None, _sent=None, _ctx=None):
            return _M_omero.model.ContrastStretchingContext._op_setYend.begin(self, ((theYend, ), _response, _ex, _sent, _ctx))

        def end_setYend(self, _r):
            return _M_omero.model.ContrastStretchingContext._op_setYend.end(self, _r)

        def checkedCast(proxy, facetOrCtx=None, _ctx=None):
            return _M_omero.model.ContrastStretchingContextPrx.ice_checkedCast(proxy, '::omero::model::ContrastStretchingContext', facetOrCtx, _ctx)
        checkedCast = staticmethod(checkedCast)

        def uncheckedCast(proxy, facet=None):
            return _M_omero.model.ContrastStretchingContextPrx.ice_uncheckedCast(proxy, facet)
        uncheckedCast = staticmethod(uncheckedCast)

        def ice_staticId():
            return '::omero::model::ContrastStretchingContext'
        ice_staticId = staticmethod(ice_staticId)

    _M_omero.model._t_ContrastStretchingContextPrx = IcePy.defineProxy('::omero::model::ContrastStretchingContext', ContrastStretchingContextPrx)

    _M_omero.model._t_ContrastStretchingContext = IcePy.declareClass('::omero::model::ContrastStretchingContext')

    _M_omero.model._t_ContrastStretchingContext = IcePy.defineClass('::omero::model::ContrastStretchingContext', ContrastStretchingContext, -1, (), True, False, _M_omero.model._t_CodomainMapContext, (), (
        ('_xstart', (), _M_omero._t_RInt, False, 0),
        ('_ystart', (), _M_omero._t_RInt, False, 0),
        ('_xend', (), _M_omero._t_RInt, False, 0),
        ('_yend', (), _M_omero._t_RInt, False, 0)
    ))
    ContrastStretchingContext._ice_type = _M_omero.model._t_ContrastStretchingContext

    ContrastStretchingContext._op_getXstart = IcePy.Operation('getXstart', Ice.OperationMode.Normal, Ice.OperationMode.Normal, False, None, (), (), (), ((), _M_omero._t_RInt, False, 0), ())
    ContrastStretchingContext._op_setXstart = IcePy.Operation('setXstart', Ice.OperationMode.Normal, Ice.OperationMode.Normal, False, None, (), (((), _M_omero._t_RInt, False, 0),), (), None, ())
    ContrastStretchingContext._op_getYstart = IcePy.Operation('getYstart', Ice.OperationMode.Normal, Ice.OperationMode.Normal, False, None, (), (), (), ((), _M_omero._t_RInt, False, 0), ())
    ContrastStretchingContext._op_setYstart = IcePy.Operation('setYstart', Ice.OperationMode.Normal, Ice.OperationMode.Normal, False, None, (), (((), _M_omero._t_RInt, False, 0),), (), None, ())
    ContrastStretchingContext._op_getXend = IcePy.Operation('getXend', Ice.OperationMode.Normal, Ice.OperationMode.Normal, False, None, (), (), (), ((), _M_omero._t_RInt, False, 0), ())
    ContrastStretchingContext._op_setXend = IcePy.Operation('setXend', Ice.OperationMode.Normal, Ice.OperationMode.Normal, False, None, (), (((), _M_omero._t_RInt, False, 0),), (), None, ())
    ContrastStretchingContext._op_getYend = IcePy.Operation('getYend', Ice.OperationMode.Normal, Ice.OperationMode.Normal, False, None, (), (), (), ((), _M_omero._t_RInt, False, 0), ())
    ContrastStretchingContext._op_setYend = IcePy.Operation('setYend', Ice.OperationMode.Normal, Ice.OperationMode.Normal, False, None, (), (((), _M_omero._t_RInt, False, 0),), (), None, ())

    _M_omero.model.ContrastStretchingContext = ContrastStretchingContext
    del ContrastStretchingContext

    _M_omero.model.ContrastStretchingContextPrx = ContrastStretchingContextPrx
    del ContrastStretchingContextPrx

# End of module omero.model

__name__ = 'omero'

# End of module omero



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
#
#   Copyright 2010 Glencoe Software, Inc. All rights reserved.
#   Use is subject to license terms supplied in LICENSE.txt
#
#

"""
  Callbacks to be used with asynchronous services. The
  ProcessCallbackI is also included in the omero.scripts
  module for backwards compatibility.
"""

from __future__ import division

from builtins import str
from future.utils import native_str
from past.utils import old_div
import Ice
import logging
import threading
import uuid

import omero
import omero.all
import omero.util.concurrency


PROC_LOG = logging.getLogger("omero.scripts.ProcessCallback")
DEL_LOG = logging.getLogger("omero.api.DeleteCallback")
CMD_LOG = logging.getLogger("omero.cmd.CmdCallback")


def adapter_and_category(adapter_or_client, category):
    if isinstance(adapter_or_client, Ice.ObjectAdapter):
        # This should be the case either when an
        # instance is created server-side or when
        # the user has passed in a category
        # explicitly. If it's missing, then we'll
        # have to throw
        if not category:
            raise omero.ClientError("No category available")
        return adapter_or_client, category
    else:
        # This is the case client-side, where an
        # omero.client instance is available.
        # If a category is passed we use that
        # (though it's unlikely that that will be useful)
        if not category:
            category = adapter_or_client.getCategory()
        return adapter_or_client.getAdapter(), category


class ProcessCallbackI(omero.grid.ProcessCallback):
    """
    Simple callback which registers itself with the given process.
    """

    FINISHED = "FINISHED"
    CANCELLED = "CANCELLED"
    KILLED = "KILLED"

    def __init__(self, adapter_or_client, process, poll=True, category=None):
        self.event = omero.util.concurrency.get_event(name="ProcessCallbackI")
        self.result = None
        self.poll = poll
        self.process = process
        self.adapter, self.category = \
            adapter_and_category(adapter_or_client, category)

        self.id = Ice.Identity(native_str(uuid.uuid4()), self.category)
        self.prx = self.adapter.add(self, self.id)  # OK ADAPTER USAGE
        self.prx = omero.grid.ProcessCallbackPrx.uncheckedCast(self.prx)
        process.registerCallback(self.prx)

    def block(self, ms):
        """
        Should only be used if the default logic of the process methods is
        kept in place. If "event.set" does not get called, this method will
        always block for the given milliseconds.
        """
        if self.poll:
            try:
                rc = self.process.poll()
                if rc is not None:
                    self.processFinished(rc.getValue())
            except Exception as e:
                PROC_LOG.warn("Error calling poll: %s" % e)

        self.event.wait(old_div(float(ms), 1000))
        if self.event.isSet():
            return self.result
        return None

    def processCancelled(self, success, current=None):
        self.result = ProcessCallbackI.CANCELLED
        self.event.set()

    def processFinished(self, returncode, current=None):
        self.result = ProcessCallbackI.FINISHED
        self.event.set()

    def processKilled(self, success, current=None):
        self.result = ProcessCallbackI.KILLED
        self.event.set()

    def close(self):
        self.adapter.remove(self.id)  # OK ADAPTER USAGE


class CmdCallbackI(omero.cmd.CmdCallback):
    """
    Callback servant used to wait until a HandlePrx would
    return non-null on getReponse. The server will notify
    of completion to prevent constantly polling on
    getResponse. Subclasses can override methods for handling
    based on the completion status.

    Example usage::

        cb = CmdCallbackI(client, handle)
        response = None
        while (response is None):
            response = cb.block(500)

        # or

        response = cb.loop(5, 500)
    """

    def __init__(self, adapter_or_client, handle, category=None,
                 foreground_poll=True):

        if adapter_or_client is None:
            raise omero.ClientError("Null client")

        if handle is None:
            raise omero.ClientError("Null handle")

        self.event = omero.util.concurrency.get_event(name="CmdCallbackI")
        self.state = (None, None)  # (Response, Status)
        self.handle = handle
        self.adapter, self.category = \
            adapter_and_category(adapter_or_client, category)

        self.id = Ice.Identity(native_str(uuid.uuid4()), self.category)
        self.prx = self.adapter.add(self, self.id)  # OK ADAPTER USAGE
        self.prx = omero.cmd.CmdCallbackPrx.uncheckedCast(self.prx)
        handle.addCallback(self.prx)
        self.initialPoll(foreground_poll)

    def initialPoll(self, foreground_poll=False):
        """
        Called at the end of construction to check a race condition.

        If HandlePrx finishes its execution before the
        CmdCallbackPrx has been sent set via addCallback,
        then there's a chance that this implementation will never
        receive a call to finished, leading to perceived hangs.

        By default, this method starts a background thread and
        calls poll(). An Ice.ObjectNotExistException
        implies that another caller has already closed the
        HandlePrx. By passing, foreground_poll=True, the poll()
        invocation can be performed in the calling thread as in
        5.1.0 and before.
        """
        if foreground_poll:
            return self.poll()

        class T(threading.Thread):

            def run(this):
                try:
                    self.poll()
                except:
                    # don't throw any exceptions, e.g. if the
                    # handle has already been closed.
                    self.onFinished(None, None, None)

        T().start()

    #
    # Local invocations
    #

    def getResponse(self):
        """
        Returns possibly null Response value. If null, then neither has
        the remote server nor the local poll method called finish
        with non-null values.
        """
        return self.state[0]

    def getStatus(self):
        """
        Returns possibly null Status value. If null, then neither has
        the remote server nor the local poll method called finish
        with non-null values.
        """
        return self.state[1]

    def getStatusOrThrow(self):
        s = self.getStatus()
        if not s:
            raise omero.ClientError("Status not present!")
        return s

    def isCancelled(self):
        """
        Returns whether Status::CANCELLED is contained in
        the flags variable of the Status instance. If no
        Status is available, a ClientError will be thrown.
        """
        s = self.getStatusOrThrow()
        try:
            s.flags.index(omero.cmd.State.CANCELLED)
            return True
        except:
            return False

    def isFailure(self):
        """
        Returns whether Status::FAILURE is contained in
        the flags variable of the Status instance. If no
        Status is available, a ClientError will be thrown.
        """
        s = self.getStatusOrThrow()
        try:
            s.flags.index(omero.cmd.State.FAILURE)
            return True
        except:
            return False

    def loop(self, loops, ms):
        """
        Calls block(long) "loops" number of times with the "ms"
        argument. This means the total wait time for the delete to occur
        is: loops X ms. Sensible values might be 10 loops for 500 ms, or
        5 seconds.

        @param loops Number of times to call block(long)
        @param ms Number of milliseconds to pass to block(long
        @throws omero.LockTimeout if block(long) does not return
        a non-null value after loops calls.
        """

        count = 0
        found = False
        while count < loops:
            count += 1
            found = self.block(ms)
            if found:
                break

        if found:
            return self.getResponse()
        else:
            waited = (old_div(ms, 1000.0)) * loops
            raise omero.LockTimeout(
                None, None, "Command unfinished after %s seconds" % waited,
                5000, int(waited))

    def block(self, ms):
        """
        Blocks for the given number of milliseconds unless
        finished(Response, Status, Current) has been called in
        which case it returns immediately with true. If false
        is returned, then the timeout was reached.
        """
        self.event.wait(old_div(float(ms), 1000))
        return self.event.isSet()

    #
    # Remote invocations
    #

    def poll(self):
        """
        Calls HandlePrx#getResponse in order to check for a
        non-null value. If so, {@link Handle#getStatus} is also called, and the
        two non-null values are passed to finished(Response, Status, Current).
        This should typically not be used. Instead, favor the use of block and
        loop.
        """
        rsp = self.handle.getResponse()
        if rsp is not None:
            s = self.handle.getStatus()
            # Only time that current should be null
            self.finished(rsp, s, None)

    def step(self, complete, total, current=None):
        """
        Called periodically by the server to signal that processing is
        moving forward. Default implementation does nothing.
        """
        pass

    def finished(self, rsp, status, current=None):
        """
        Called when the command has completed whether with
        a cancellation or a completion.
        """
        self.state = (rsp, status)
        self.event.set()
        self.onFinished(rsp, status, current)

    def onFinished(self, rsp, status, current):
        """
        Method intended to be overridden by subclasses. Default logic does
        nothing.
        """
        pass

    def close(self, closeHandle):
        """
        First removes self from the adapter so as to no longer receive
        notifications, and the calls close on the remote handle if requested.
        """
        self.adapter.remove(self.id)  # OK ADAPTER USAGE
        if closeHandle:
            self.handle.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# Copyright (C) 2009-2016 Glencoe Software, Inc. All Rights Reserved.
# Use is subject to license terms supplied in LICENSE.txt
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
   Startup plugin for command-line importer.

"""
from __future__ import division
from __future__ import print_function

from builtins import str
from past.utils import old_div
from past.builtins import basestring
from builtins import object
from io import BytesIO
import os
import csv
import sys
import shlex
import requests
import re
from zipfile import ZipFile


from omero.cli import BaseControl, CLI
import omero.java
from omero.util import get_omero_user_cache_dir
from omero_ext.argparse import SUPPRESS
from omero_ext.path import path

START_CLASS = "ome.formats.importer.cli.CommandLineImporter"
TEST_CLASS = "ome.formats.test.util.TestEngine"

HELP = """Run the Java-based command-line importer

This is a Python wrapper around the Java importer. Login is handled by Python
OMERO.cli. To see more options, use "--javahelp".

Options marked with "**" are passed strictly to Java. If they interfere with
any of the Python arguments, you may need to end precede your arguments with a
"--".

Bulk imports:

Rather than passing one or more files to the import command, a single
dictionary-like file (e.g. yml or json) can be passed to the `--bulk`
argument. Most keys in the bulk file will be treated like additional
command-line arguments. Special keys include:

 * columns      A list of columns for parsing the value of path
 * continue     Like the "-c" changes error handling
 * dry_run      If true, print out additional arguments rather than run them.
                If another string other than false, use as a template for
                storing the import commands. (e.g. /tmp/%s.sh)
 * include      Relative path (from the bulk file) of a parent bulk file
 * path         A file which will be parsed line by line based on its file
                ending. Lines containing zero or more keys along with a
                single file to be imported. Options for formats include:
                  - .tsv and .csv files will be parsed by the existing library
                  - other files will be parsed with shlex
                  - unless no columns are specified, in which case each line
                    is treated as a file

"""
EXAMPLES = """
Examples:

  # Display help
  $ omero import -h
  # Import foo.tiff using current login
  $ omero import ~/Data/my_file.dv
  # Import foo.tiff using input credentials
  $ omero import -s localhost -u user -w password foo.tiff
  # Set Java debugging level to ALL
  $ omero import foo.tiff -- --debug=ALL
  # Display used files for importing foo.tiff
  $ omero import foo.tiff -f
  # Limit debugging output
  $ omero import -- --debug=ERROR foo.tiff

For additional information, see:
https://docs.openmicroscopy.org/latest/omero/users/cli/import.html
Report bugs at https://forum.image.sc/
"""
TESTHELP = """Run the Importer TestEngine suite (devs-only)"""
DEBUG_CHOICES = ["ALL", "DEBUG", "ERROR", "FATAL", "INFO", "TRACE", "WARN"]
OUTPUT_CHOICES = ["ids", "legacy", "yaml"]
SKIP_CHOICES = ['all', 'checksum', 'minmax', 'thumbnails', 'upgrade']
NO_ARG = object()

OMERO_JAVA_ZIP = (
    'https://downloads.openmicroscopy.org/omero/{version}/OMERO.java.zip'
)


class CommandArguments(object):

    def __init__(self, ctx, args):
        self.__ctx = ctx
        self.__args = args
        self.__accepts = set()
        self.__added = dict()
        self.__java_initial = list()
        self.__java_additional = list()
        self.__py_initial = list()
        self.__py_additional = list()
        # Python arguments
        self.__py_keys = (
            "javahelp", "skip", "file", "errs", "logback",
            "port", "password", "group", "create", "func",
            "bulk", "prog", "user", "key", "path", "logprefix",
            "JAVA_DEBUG", "quiet", "server", "depth", "clientdir",
            "fetch_jars",
            "sudo")
        self.set_login_arguments(ctx, args)
        self.set_skip_arguments(args)

        for key in vars(args):
            self.__accepts.add(key)
            val = getattr(args, key)
            if key in self.__py_keys:
                # Place the Python elements on the CommandArguments
                # instance so that it behaves like `args`
                setattr(self, key, val)
                self.append_arg(self.__py_initial, key, val)

            elif not val:
                # If there's no value, do nothing
                pass

            else:
                self.append_arg(self.__java_initial, key, val)

    def append_arg(self, cmd_list, key, val=NO_ARG):
        arg_list = self.build_arg_list(key, val)
        cmd_list.extend(arg_list)

    def reset_arg(self, cmd_list, idx, key, val=NO_ARG):
        arg_list = self.build_arg_list(key, val)
        cmd_list[idx:idx+len(arg_list)] = arg_list

    def build_arg_list(self, key, val=NO_ARG):
        arg_list = []
        if len(key) == 1:
            arg_list.append("-"+key)
            if val != NO_ARG:
                if isinstance(val, basestring):
                    arg_list.append(val)
        else:
            key = key.replace("_", "-")
            if val == NO_ARG:
                arg_list.append("--%s" % key)
            elif isinstance(val, basestring):
                arg_list.append(
                    "--%s=%s" % (key, val))
            else:
                arg_list.append("--%s" % key)
        return arg_list

    def set_path(self, path):
        if not isinstance(path, list):
            self.__ctx.die(202, "Path is not a list")
        else:
            self.path = path

    def java_args(self):
        rv = list()
        rv.extend(self.__java_initial)
        rv.extend(self.__java_additional)
        rv.extend(self.path)
        if self.JAVA_DEBUG:
            # Since "args.debug" is used by omero/cli.py itself,
            # uses of "--debug" *after* the `import` command are
            # handled by placing them in this special variable.
            rv.append("--debug=%s" % self.JAVA_DEBUG)
        return rv

    def initial_args(self):
        rv = list()
        rv.extend(self.__py_initial)
        rv.extend(self.__java_initial)
        return rv

    def added_args(self):
        rv = list()
        rv.extend(self.__py_additional)
        rv.extend(self.__java_additional)
        rv.extend(self.path)
        return rv

    def accepts(self, key):
        return key in self.__accepts

    def add(self, key, val=NO_ARG):

        idx = None
        if key in self.__added:
            idx = self.__added[key]

        if key in self.__py_keys:
            # First we check if this is a Python argument, in which
            # case it's set directly on the instance itself. This
            # may need to be later set elsewhere if multiple bulk
            # files are supported.
            setattr(self, key, val)
            where = self.__py_additional
        elif not self.accepts(key):
            self.__ctx.die(200, "Unknown argument: %s" % key)
        else:
            where = self.__java_additional

        if idx is None:
            idx = len(where)
            self.append_arg(where, key, val)
            self.__added[key] = idx
        else:
            self.reset_arg(where, idx, key, val)

    def set_login_arguments(self, ctx, args):
        """Set the connection arguments"""

        if args.javahelp:
            self.__java_initial.append("-h")

        # Connection is required unless help arguments or -f is passed
        connection_required = ("-h" not in self.__java_initial and
                               not args.f and
                               not args.advanced_help)
        if connection_required:
            client = ctx.conn(args)
            host = client.getProperty("omero.host")
            port = client.getProperty("omero.port")
            session = client.getSessionId()
            self.__java_initial.extend(["-s", host])
            self.__java_initial.extend(["-p", port])
            self.__java_initial.extend(["-k", session])

    def set_skip_arguments(self, args):
        """Set the arguments to skip steps during import"""
        if not args.skip:
            return
        self.set_skip_values(args.skip)

    def set_skip_values(self, skip):
        """Set the arguments to skip steps during import"""

        if ('all' in skip or 'checksum' in skip):
            self.__java_initial.append("--checksum-algorithm=File-Size-64")
        if ('all' in skip or 'thumbnails' in skip):
            self.__java_initial.append("--no-thumbnails")
        if ('all' in skip or 'minmax' in skip):
            self.__java_initial.append("--no-stats-info")
        if ('all' in skip or 'upgrade' in skip):
            self.__java_initial.append("--no-upgrade-check")

    def open_files(self, mode="w"):
        # Open file handles for stdout/stderr if applicable
        out = self.open_log(self.__args.file, self.__args.logprefix, mode=mode)
        err = self.open_log(self.__args.errs, self.__args.logprefix, mode=mode)
        return out, err

    def open_log(self, file, prefix=None, mode="w"):
        if not file:
            return None
        if prefix:
            file = os.path.sep.join([prefix, file])
        file = os.path.abspath(file)
        dir = os.path.dirname(file)
        if not os.path.exists(dir):
            os.makedirs(dir)
        return open(file, mode)


class ImportControl(BaseControl):

    COMMAND = [START_CLASS]

    def _configure(self, parser):

        parser.add_login_arguments()

        parser.add_argument(
            "--javahelp", "--java-help",
            action="store_true", help="Show the Java help text")

        # The following arguments are strictly used by Python
        # The "---" form is kept for backwards compatibility.
        py_group = parser.add_argument_group(
            'Python arguments',
            'Optional arguments which are used to configure import.')

        def add_python_argument(*args, **kwargs):
            py_group.add_argument(*args, **kwargs)

        for name, help in (
            ("bulk", "Bulk YAML file for driving multiple imports"),
            ("logprefix", "Directory or file prefix for --file and --errs"),
            (
                "file",
                "File for storing the standard output from the Java process"
            ),
            (
                "errs",
                "File for storing the standard error from the Java process"
            )
        ):
            add_python_argument("--%s" % name, nargs="?", help=help)
            add_python_argument("---%s" % name, nargs="?", help=SUPPRESS)

        add_python_argument(
            "--clientdir", type=str,
            help="Path to the directory containing the client JARs. "
            " Default: lib/client")
        add_python_argument(
            "--logback", type=str,
            help="Path to a logback xml file. "
            " Default: etc/logback-cli.xml")
        add_python_argument(
            "--fetch-jars", type=str,
            help="Download OMERO.java jars by version or URL, then exit")

        # The following arguments are strictly passed to Java
        name_group = parser.add_argument_group(
            'Naming arguments', 'Optional arguments passed strictly to Java.')

        def add_java_name_argument(*args, **kwargs):
            name_group.add_argument(*args, **kwargs)

        add_java_name_argument(
            "-n", "--name",
            help="Image or plate name to use (**)",
            metavar="NAME")
        add_java_name_argument(
            "-x", "--description",
            help="Image or plate description to use (**)",
            metavar="DESCRIPTION")
        # Deprecated naming arguments
        add_java_name_argument(
            "--plate_name",
            help=SUPPRESS)
        add_java_name_argument(
            "--plate_description",
            help=SUPPRESS)

        # Feedback options
        feedback_group = parser.add_argument_group(
            'Feedback arguments',
            'Optional arguments passed strictly to Java allowing to report'
            ' errors to the OME team.')

        def add_feedback_argument(*args, **kwargs):
            feedback_group.add_argument(*args, **kwargs)

        add_feedback_argument(
            "--report", action="store_true",
            help="Report errors to the OME team (**)")
        add_feedback_argument(
            "--upload", action="store_true",
            help=("Upload broken files and log file (if any) with report."
                  " Required --report (**)"))
        add_feedback_argument(
            "--logs", action="store_true",
            help=("Upload log file (if any) with report."
                  " Required --report (**)"))
        add_feedback_argument(
            "--email",
            help="Email for reported errors. Required --report (**)",
            metavar="EMAIL")
        add_feedback_argument(
            "--qa-baseurl",
            help=SUPPRESS)

        # Annotation options
        annotation_group = parser.add_argument_group(
            'Annotation arguments',
            'Optional arguments passed strictly to Java allowing to annotate'
            ' imports.')

        def add_annotation_argument(*args, **kwargs):
            annotation_group.add_argument(*args, **kwargs)

        add_annotation_argument(
            "--annotation-ns", metavar="ANNOTATION_NS",
            help="Namespace to use for subsequent annotation (**)")
        add_annotation_argument(
            "--annotation-text", metavar="ANNOTATION_TEXT",
            help="Content for a text annotation (**)")
        add_annotation_argument(
            "--annotation-link",
            metavar="ANNOTATION_LINK",
            help="Comment annotation ID to link all images to (**)")
        add_annotation_argument(
            "--annotation_ns", metavar="ANNOTATION_NS",
            help=SUPPRESS)
        add_annotation_argument(
            "--annotation_text", metavar="ANNOTATION_TEXT",
            help=SUPPRESS)
        add_annotation_argument(
            "--annotation_link", metavar="ANNOTATION_LINK",
            help=SUPPRESS)

        java_group = parser.add_argument_group(
            'Java arguments', 'Optional arguments passed strictly to Java.')

        def add_java_argument(*args, **kwargs):
            java_group.add_argument(*args, **kwargs)

        add_java_argument(
            "-f", action="store_true",
            help="Display the used files and exit (**)")
        add_java_argument(
            "-c", action="store_true",
            help="Continue importing after errors (**)")
        add_java_argument(
            "-l", "--readers",
            help="Use the list of readers rather than the default (**)",
            metavar="READER_FILE")
        add_java_argument(
            "-d",
            help="OMERO dataset ID to import image into (**)",
            metavar="DATASET_ID")
        add_java_argument(
            "-r",
            help="OMERO screen ID to import plate into (**)",
            metavar="SCREEN_ID")
        add_java_argument(
            "-T", "--target",
            help="OMERO target specification (**)",
            metavar="TARGET")
        add_java_argument(
            "--debug", choices=DEBUG_CHOICES,
            help="Turn debug logging on (**)",
            metavar="LEVEL", dest="JAVA_DEBUG")
        add_java_argument(
            "--output", choices=OUTPUT_CHOICES,
            help="Set an alternative output style",
            metavar="TYPE")
        add_java_argument(
            "--encrypted",
            choices=("true", "false"),
            help="Whether the import should use SSL or not",
            metavar="TYPE")

        # Arguments previously *following" `--`
        advjava_group = parser.add_argument_group(
            'Advanced Java arguments', (
                'Optional arguments passed strictly to Java. '
                'For more information, see --advanced-help.'))

        def add_advjava_argument(*args, **kwargs):
            advjava_group.add_argument(*args, **kwargs)

        add_advjava_argument(
            "--advanced-help", action="store_true",
            help="Show the advanced help text")
        add_advjava_argument(
            "--transfer", nargs="?", metavar="TYPE",
            help="Transfer methods like in-place import")
        add_advjava_argument(
            "--exclude", nargs="?", metavar="TYPE",
            help="Exclusion filters for preventing re-import")
        add_advjava_argument(
            "--checksum-algorithm", nargs="?", metavar="TYPE",
            help="Alternative hashing mechanisms balancing speed & accuracy")
        add_advjava_argument(
            "--no-stats-info", action="store_true", help=SUPPRESS)
        add_advjava_argument(
            "--no-thumbnails", action="store_true", help=SUPPRESS)
        add_advjava_argument(
            "--no-upgrade-check", action="store_true", help=SUPPRESS)
        add_advjava_argument(
            "--parallel-upload", metavar="COUNT",
            help="Number of file upload threads to run at the same time")
        add_advjava_argument(
            "--parallel-fileset", metavar="COUNT",
            help="Number of fileset candidates to import at the same time")

        # Unsure on these.
        add_python_argument(
            "--depth", default=4, type=int,
            help="Number of directories to scan down for files")
        add_python_argument(
            "--skip", type=str, choices=SKIP_CHOICES, action='append',
            help="Optional step to skip during import")
        add_python_argument(
            "path", nargs="*",
            help="Path to be passed to the Java process")

        parser.set_defaults(func=self.importer)

    def _get_classpath_logback(self, args):
        lib_client = self.ctx.dir / "lib" / "client"
        auto_download = False
        if args.clientdir:
            client_dir = path(args.clientdir)
        elif lib_client.exists():
            client_dir = lib_client
        else:
            auto_download = True
            omero_java_dir, omero_java_txt = self._userdir_jars()
            client_dir = omero_java_dir

        etc_dir = old_div(self.ctx.dir, "etc")
        if args.logback:
            xml_file = path(args.logback)
        else:
            xml_file = old_div(etc_dir, "logback-cli.xml")

        classpath = []
        if client_dir and client_dir.exists():
            classpath = [f.abspath() for f in client_dir.files("*.jar")]
        if auto_download:
            if classpath:
                self.ctx.err('Using {}'.format(omero_java_txt.text()))
                if not args.logback:
                    xml_file = client_dir / "logback-cli.xml"
        else:
            if not classpath:
                self.ctx.die(
                    103, "No JAR files found under '%s'" % client_dir)

        logback = "-Dlogback.configurationFile=%s" % xml_file
        return classpath, logback

    def importer(self, args):
        if args.fetch_jars:
            if args.path:
                self.ctx.err('WARNING: Ignoring extra arguments')
            self.download_omero_java(args.fetch_jars)
            return

        classpath, logback = self._get_classpath_logback(args)
        if not classpath:
            self.download_omero_java('latest')
            classpath, logback = self._get_classpath_logback(args)

        command_args = CommandArguments(self.ctx, args)
        xargs = [logback, "-Xmx1024M", "-cp", os.pathsep.join(classpath)]
        xargs.append("-Domero.import.depth=%s" % args.depth)

        if args.bulk and args.path:
            self.ctx.die(104, "When using bulk import, omit paths")
        elif args.bulk:
            self.bulk_import(command_args, xargs)
        else:
            self.do_import(command_args, xargs)

    def _userdir_jars(self, parentonly=False):
        user_jars = get_omero_user_cache_dir() / 'jars'
        # Use this file instead of a symlink so it works on all platform
        omero_java_txt = user_jars / 'OMERO.java.txt'
        if parentonly:
            return user_jars, omero_java_txt
        omero_java_dir = None
        if omero_java_txt.exists():
            omero_java_dir = omero_java_txt.text().strip()
            return user_jars / omero_java_dir / 'libs', omero_java_txt
        else:
            return None, omero_java_txt

    def download_omero_java(self, version_or_uri):
        if re.match("^\w+://", version_or_uri):
            omero_java_zip = version_or_uri
        else:
            omero_java_zip = OMERO_JAVA_ZIP.format(version=version_or_uri)
        self.ctx.err("Downloading %s" % omero_java_zip)
        jars_dir, omero_java_txt = self._userdir_jars(parentonly=True)
        jars_dir.makedirs_p()
        with requests.get(omero_java_zip) as resp:
            with ZipFile(BytesIO(resp.content)) as zipfile:
                topdirs = set(f.filename.split("/")[0]
                              for f in zipfile.filelist if f.is_dir())
                if len(topdirs) != 1:
                    self.ctx.die(
                        108,
                        'Expected one top directory in OMERO.java.zip: {}'
                        .format(topdirs))
                topdir = topdirs.pop()
                if os.path.isabs(topdir):
                    self.ctx.die(
                        108,
                        'Unexpected absolute paths in OMERO.java.zip: {}'
                        .format(topdir))
                zipfile.extractall(jars_dir)
                omero_java_txt.write_text(topdir)

    def do_import(self, command_args, xargs, mode="w"):
        out = err = None
        try:

            import_command = self.COMMAND + command_args.java_args()
            out, err = command_args.open_files(mode=mode)

            p = omero.java.popen(
                import_command, debug=False, xargs=xargs,
                stdout=out, stderr=err)

            self.ctx.rv = p.wait()

        finally:
            # Make sure file handles are closed
            if out:
                out.close()
            if err:
                err.close()

    def bulk_import(self, command_args, xargs):

        try:
            from yaml import safe_load
        except ImportError:
            self.ctx.die(105, "ERROR: PyYAML is not installed")

        old_pwd = os.getcwd()
        try:

            # Walk the .yml graph looking for includes
            # and load them all so that the top parent
            # values can be overwritten.
            contents = list()
            bulkfile = command_args.bulk
            while bulkfile:
                bulkfile = os.path.abspath(bulkfile)
                parent = os.path.dirname(bulkfile)
                with open(bulkfile, "r") as f:
                    data = safe_load(f)
                    contents.append((bulkfile, parent, data))
                    bulkfile = data.get("include")
                    os.chdir(parent)
                    # TODO: included files are updated based on the including
                    # file but other file paths aren't!

            bulk = dict()
            for bulkfile, parent, data in reversed(contents):
                bulk.update(data)
                os.chdir(parent)

            incr = 0
            failed = 0
            total = 0
            for cont in self.parse_bulk(bulk, command_args):
                incr += 1
                if command_args.dry_run:
                    rv = ['"%s"' % x for x in command_args.added_args()]
                    rv = " ".join(rv)
                    if command_args.dry_run.lower() == "true":
                        self.ctx.out(rv)
                    else:
                        with open(command_args.dry_run % incr, "w") as o:
                            # FIXME: this assumes 'omero'
                            print(sys.argv[0], "import", rv, file=o)
                else:
                    if incr == 1:
                        mode = "w"
                    else:
                        mode = "a"
                    self.do_import(command_args, xargs, mode=mode)
                if self.ctx.rv:
                    failed += 1
                    total += self.ctx.rv
                    if cont:
                        msg = "Import failed with error code: %s. Continuing"
                        self.ctx.err(msg % self.ctx.rv)
                    else:
                        msg = "Import failed. Use -c to continue after errors"
                        self.ctx.die(106, msg)
                # Fail if any import failed
                self.ctx.rv = total
                if failed:
                    self.ctx.err("%x failed imports" % failed)
        finally:
            os.chdir(old_pwd)

    def parse_bulk(self, bulk, command_args):
        # Known keys with special handling
        cont = False

        command_args.dry_run = False
        if "dry_run" in bulk:
            dry_run = str(bulk.pop("dry_run"))
            # Accept any non-false string since it might be a pattern
            if dry_run.lower() != "false":
                command_args.dry_run = dry_run

        if "continue" in bulk:
            cont = True
            c = bulk.pop("continue")
            if bool(c):
                command_args.add("c")

        if "skip" in bulk:
            command_args.set_skip_values(bulk.pop("skip"))

        if "path" not in bulk:
            # Required until @file format is implemented
            self.ctx.die(107, "No path specified")
        path = bulk.pop("path")

        cols = None
        if "columns" in bulk:
            cols = bulk.pop("columns")

        if "include" in bulk:
            bulk.pop("include")

        # Now parse all other keys
        for key in bulk:
            command_args.add(key, bulk[key])

        # All properties are set, yield for each path
        # to be imported in turn. The value for `cont`
        # is yielded so that the caller knows whether
        # or not an error should be fatal.

        if not cols:
            # No parsing necessary
            function = self.parse_text
        else:
            function = self.parse_shlex
            if path.endswith(".tsv"):
                function = self.parse_tsv
            elif path.endswith(".csv"):
                function = self.parse_csv

        for parts in function(path):
            if not cols:
                command_args.set_path(parts)
            else:
                for idx, col in enumerate(cols):
                    if col == "path":
                        command_args.set_path([parts[idx]])
                    else:
                        command_args.add(col, parts[idx])
            yield cont

    def parse_text(self, path, parse=False):
        with open(path, "r") as o:
            for line in o:
                line = line.strip()
                if parse:
                    line = shlex.split(line)
                yield [line]

    def parse_shlex(self, path):
        for line in self.parse_text(path, parse=True):
            yield line

    def parse_tsv(self, path, delimiter="\t"):
        for line in self.parse_csv(path, delimiter):
            yield line

    def parse_csv(self, path, delimiter=","):
        with open(path, "r") as data:
            for line in csv.reader(data, delimiter=delimiter):
                yield line


class TestEngine(ImportControl):
    COMMAND = [TEST_CLASS]


try:
    register("import", ImportControl, HELP, epilog=EXAMPLES)
    register("testengine", TestEngine, TESTHELP)
except NameError:
    if __name__ == "__main__":
        cli = CLI()
        cli.register("import", ImportControl, HELP, epilog=EXAMPLES)
        cli.register("testengine", TestEngine, TESTHELP)
        cli.invoke(sys.argv[1:])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# Copyright 2009 Glencoe Software, Inc. All rights reserved.
# Use is subject to license terms supplied in LICENSE.txt

"""
   Startup plugin for command-line exporter
"""

from builtins import str
import os
import sys

from omero.cli import BaseControl, CLI, NewFileType

HELP = """Support for exporting data in XML and TIFF formats

Example Usage:
  omero export --file new.ome.tif            Image:1
  omero export --file new.ome.xml --type XML Image:1
  omero export --file output-dir  --iterate  Dataset:2

If --iterate is used, the only supported obj is Dataset:<id>
"""


# Default block read size when downloading exported data
DEFAULT_READ_LENGTH = 1000*1000


class ExportControl(BaseControl):

    def _configure(self, parser):
        parser.add_argument(
            "-f", "--file", type=NewFileType("wb"), required=True,
            help="Filename to export to or '-' for stdout."
            " File may not exist")
        parser.add_argument(
            "-t", "--type", default="TIFF", choices=("TIFF", "XML"),
            help="Type of export. Default: %(default)s")
        parser.add_argument("obj", help="Format: Image:<id>")
        parser.add_argument(
            "--iterate", action="store_true", default=False,
            help="Iterate over an object and write individual objects to the"
            " directory named by --file (EXPERIMENTAL)")

        parser.set_defaults(func=self.export)
        parser.add_login_arguments()

    def export(self, args):

        img = args.obj
        if 0 > img.find(":"):
            self.ctx.die(5, "Format: 'Image:<id>'")

        klass, id = img.split(":")
        try:
            id = int(id)
        except:
            self.ctx.die(3, "Bad id format: %s" % id)

        images = []
        datasets = []

        if klass == "Image":
            images.append(id)
            self.handleImages(args, images)
        elif klass == "Dataset":
            if not args.iterate:
                self.ctx.die(4, "Dataset currently only supported with"
                                " --iterate")
            datasets.append(id)
            self.handleDatasets(args, datasets)
        else:
            self.ctx.die(5, "Can't add type: %s" % klass)

    def handleDatasets(self, args, datasets):

        if args.file == sys.stdout:
            self.ctx.die(950, "Can't use stdin for datasets")

        f = args.file
        dir = f.name
        f.close()
        os.remove(dir)
        os.makedirs(dir)

        c = self.ctx.conn(args)

        import omero
        p = omero.sys.ParametersI()
        p.leaves()

        ds = c.sf.getContainerService().loadContainerHierarchy("Dataset",
                                                               datasets, p)
        if not ds:
            self.ctx.die(7, "No datasets found: %s",
                         ", ".join([str(x) for x in datasets]))

        for d in ds:
            for i in d.linkedImageList():
                if i:
                    i = i.id.val
                    args.file = open(
                        os.path.join(dir, "%s.ome.%s"
                                     % (i, args.type.lower())), "wb")
                    self.handleImages(args, [i])

    def handleImages(self, args, images):
        e = None
        handle = args.file
        issysout = (handle == sys.stdout)

        c = self.ctx.conn(args)
        e = c.getSession().createExporter()

        remove = True
        try:
            for img in images:
                e.addImage(img)

            import omero
            try:
                if args.type == "TIFF":
                    l = e.generateTiff()
                else:
                    l = e.generateXml()
            except omero.ServerError as se:
                self.ctx.err("%s: %s" % (se.__class__.__name__, se.message))
                return

            remove = False
            offset = 0
            while True:
                rv = e.read(offset, DEFAULT_READ_LENGTH)
                if not rv:
                    break
                rv = rv[:min(DEFAULT_READ_LENGTH, l - offset)]
                offset += len(rv)
                handle.write(rv)

        finally:
            try:
                if not issysout:
                    handle.close()
                    if remove:
                        os.remove(handle.name)
            except Exception as e:
                self.ctx.err("Failed to close handle: %s" % e)

            e.close()

try:
    register("export", ExportControl, HELP)
except NameError:
    if __name__ == "__main__":
        cli = CLI()
        cli.register("export", ExportControl, HELP)
        cli.invoke(sys.argv[1:])



# -*- coding: utf-8 -*-
# **********************************************************************
#
# Copyright (c) 2003-2017 ZeroC, Inc. All rights reserved.
#
# This copy of Ice is licensed to you under the terms described in the
# ICE_LICENSE file included in this distribution.
#
# **********************************************************************
#
# Ice version 3.6.4
#
# <auto-generated>
#
# Generated from file `Basic.ice'
#
# Warning: do not edit this file.
#
# </auto-generated>
#

from sys import version_info as _version_info_
import Ice, IcePy
import omero_cmd_API_ice

# Included module omero
_M_omero = Ice.openModule('omero')

# Included module Ice
_M_Ice = Ice.openModule('Ice')

# Included module Glacier2
_M_Glacier2 = Ice.openModule('Glacier2')

# Included module omero.model
_M_omero.model = Ice.openModule('omero.model')

# Included module omero.sys
_M_omero.sys = Ice.openModule('omero.sys')

# Included module omero.api
_M_omero.api = Ice.openModule('omero.api')

# Included module omero.cmd
_M_omero.cmd = Ice.openModule('omero.cmd')

# Start of module omero
__name__ = 'omero'

# Start of module omero.cmd
__name__ = 'omero.cmd'

if 'DoAll' not in _M_omero.cmd.__dict__:
    _M_omero.cmd.DoAll = Ice.createTempClass()
    class DoAll(_M_omero.cmd.Request):
        def __init__(self, requests=None, contexts=None):
            _M_omero.cmd.Request.__init__(self)
            self.requests = requests
            self.contexts = contexts

        def ice_ids(self, current=None):
            return ('::Ice::Object', '::omero::cmd::DoAll', '::omero::cmd::Request')

        def ice_id(self, current=None):
            return '::omero::cmd::DoAll'

        def ice_staticId():
            return '::omero::cmd::DoAll'
        ice_staticId = staticmethod(ice_staticId)

        def __str__(self):
            return IcePy.stringify(self, _M_omero.cmd._t_DoAll)

        __repr__ = __str__

    _M_omero.cmd.DoAllPrx = Ice.createTempClass()
    class DoAllPrx(_M_omero.cmd.RequestPrx):

        def checkedCast(proxy, facetOrCtx=None, _ctx=None):
            return _M_omero.cmd.DoAllPrx.ice_checkedCast(proxy, '::omero::cmd::DoAll', facetOrCtx, _ctx)
        checkedCast = staticmethod(checkedCast)

        def uncheckedCast(proxy, facet=None):
            return _M_omero.cmd.DoAllPrx.ice_uncheckedCast(proxy, facet)
        uncheckedCast = staticmethod(uncheckedCast)

        def ice_staticId():
            return '::omero::cmd::DoAll'
        ice_staticId = staticmethod(ice_staticId)

    _M_omero.cmd._t_DoAllPrx = IcePy.defineProxy('::omero::cmd::DoAll', DoAllPrx)

    _M_omero.cmd._t_DoAll = IcePy.declareClass('::omero::cmd::DoAll')

    _M_omero.cmd._t_DoAll = IcePy.defineClass('::omero::cmd::DoAll', DoAll, -1, (), False, False, _M_omero.cmd._t_Request, (), (
        ('requests', (), _M_omero.cmd._t_RequestList, False, 0),
        ('contexts', (), _M_omero.cmd._t_StringMapList, False, 0)
    ))
    DoAll._ice_type = _M_omero.cmd._t_DoAll

    _M_omero.cmd.DoAll = DoAll
    del DoAll

    _M_omero.cmd.DoAllPrx = DoAllPrx
    del DoAllPrx

if 'DoAllRsp' not in _M_omero.cmd.__dict__:
    _M_omero.cmd.DoAllRsp = Ice.createTempClass()
    class DoAllRsp(_M_omero.cmd.OK):
        def __init__(self, responses=None, status=None):
            _M_omero.cmd.OK.__init__(self)
            self.responses = responses
            self.status = status

        def ice_ids(self, current=None):
            return ('::Ice::Object', '::omero::cmd::DoAllRsp', '::omero::cmd::OK', '::omero::cmd::Response')

        def ice_id(self, current=None):
            return '::omero::cmd::DoAllRsp'

        def ice_staticId():
            return '::omero::cmd::DoAllRsp'
        ice_staticId = staticmethod(ice_staticId)

        def __str__(self):
            return IcePy.stringify(self, _M_omero.cmd._t_DoAllRsp)

        __repr__ = __str__

    _M_omero.cmd.DoAllRspPrx = Ice.createTempClass()
    class DoAllRspPrx(_M_omero.cmd.OKPrx):

        def checkedCast(proxy, facetOrCtx=None, _ctx=None):
            return _M_omero.cmd.DoAllRspPrx.ice_checkedCast(proxy, '::omero::cmd::DoAllRsp', facetOrCtx, _ctx)
        checkedCast = staticmethod(checkedCast)

        def uncheckedCast(proxy, facet=None):
            return _M_omero.cmd.DoAllRspPrx.ice_uncheckedCast(proxy, facet)
        uncheckedCast = staticmethod(uncheckedCast)

        def ice_staticId():
            return '::omero::cmd::DoAllRsp'
        ice_staticId = staticmethod(ice_staticId)

    _M_omero.cmd._t_DoAllRspPrx = IcePy.defineProxy('::omero::cmd::DoAllRsp', DoAllRspPrx)

    _M_omero.cmd._t_DoAllRsp = IcePy.declareClass('::omero::cmd::DoAllRsp')

    _M_omero.cmd._t_DoAllRsp = IcePy.defineClass('::omero::cmd::DoAllRsp', DoAllRsp, -1, (), False, False, _M_omero.cmd._t_OK, (), (
        ('responses', (), _M_omero.cmd._t_ResponseList, False, 0),
        ('status', (), _M_omero.cmd._t_StatusList, False, 0)
    ))
    DoAllRsp._ice_type = _M_omero.cmd._t_DoAllRsp

    _M_omero.cmd.DoAllRsp = DoAllRsp
    del DoAllRsp

    _M_omero.cmd.DoAllRspPrx = DoAllRspPrx
    del DoAllRspPrx

if 'ListRequests' not in _M_omero.cmd.__dict__:
    _M_omero.cmd.ListRequests = Ice.createTempClass()
    class ListRequests(_M_omero.cmd.Request):
        def __init__(self):
            _M_omero.cmd.Request.__init__(self)

        def ice_ids(self, current=None):
            return ('::Ice::Object', '::omero::cmd::ListRequests', '::omero::cmd::Request')

        def ice_id(self, current=None):
            return '::omero::cmd::ListRequests'

        def ice_staticId():
            return '::omero::cmd::ListRequests'
        ice_staticId = staticmethod(ice_staticId)

        def __str__(self):
            return IcePy.stringify(self, _M_omero.cmd._t_ListRequests)

        __repr__ = __str__

    _M_omero.cmd.ListRequestsPrx = Ice.createTempClass()
    class ListRequestsPrx(_M_omero.cmd.RequestPrx):

        def checkedCast(proxy, facetOrCtx=None, _ctx=None):
            return _M_omero.cmd.ListRequestsPrx.ice_checkedCast(proxy, '::omero::cmd::ListRequests', facetOrCtx, _ctx)
        checkedCast = staticmethod(checkedCast)

        def uncheckedCast(proxy, facet=None):
            return _M_omero.cmd.ListRequestsPrx.ice_uncheckedCast(proxy, facet)
        uncheckedCast = staticmethod(uncheckedCast)

        def ice_staticId():
            return '::omero::cmd::ListRequests'
        ice_staticId = staticmethod(ice_staticId)

    _M_omero.cmd._t_ListRequestsPrx = IcePy.defineProxy('::omero::cmd::ListRequests', ListRequestsPrx)

    _M_omero.cmd._t_ListRequests = IcePy.defineClass('::omero::cmd::ListRequests', ListRequests, -1, (), False, False, _M_omero.cmd._t_Request, (), ())
    ListRequests._ice_type = _M_omero.cmd._t_ListRequests

    _M_omero.cmd.ListRequests = ListRequests
    del ListRequests

    _M_omero.cmd.ListRequestsPrx = ListRequestsPrx
    del ListRequestsPrx

if 'ListRequestsRsp' not in _M_omero.cmd.__dict__:
    _M_omero.cmd.ListRequestsRsp = Ice.createTempClass()
    class ListRequestsRsp(_M_omero.cmd.OK):
        def __init__(self, list=None):
            _M_omero.cmd.OK.__init__(self)
            self.list = list

        def ice_ids(self, current=None):
            return ('::Ice::Object', '::omero::cmd::ListRequestsRsp', '::omero::cmd::OK', '::omero::cmd::Response')

        def ice_id(self, current=None):
            return '::omero::cmd::ListRequestsRsp'

        def ice_staticId():
            return '::omero::cmd::ListRequestsRsp'
        ice_staticId = staticmethod(ice_staticId)

        def __str__(self):
            return IcePy.stringify(self, _M_omero.cmd._t_ListRequestsRsp)

        __repr__ = __str__

    _M_omero.cmd.ListRequestsRspPrx = Ice.createTempClass()
    class ListRequestsRspPrx(_M_omero.cmd.OKPrx):

        def checkedCast(proxy, facetOrCtx=None, _ctx=None):
            return _M_omero.cmd.ListRequestsRspPrx.ice_checkedCast(proxy, '::omero::cmd::ListRequestsRsp', facetOrCtx, _ctx)
        checkedCast = staticmethod(checkedCast)

        def uncheckedCast(proxy, facet=None):
            return _M_omero.cmd.ListRequestsRspPrx.ice_uncheckedCast(proxy, facet)
        uncheckedCast = staticmethod(uncheckedCast)

        def ice_staticId():
            return '::omero::cmd::ListRequestsRsp'
        ice_staticId = staticmethod(ice_staticId)

    _M_omero.cmd._t_ListRequestsRspPrx = IcePy.defineProxy('::omero::cmd::ListRequestsRsp', ListRequestsRspPrx)

    _M_omero.cmd._t_ListRequestsRsp = IcePy.declareClass('::omero::cmd::ListRequestsRsp')

    _M_omero.cmd._t_ListRequestsRsp = IcePy.defineClass('::omero::cmd::ListRequestsRsp', ListRequestsRsp, -1, (), False, False, _M_omero.cmd._t_OK, (), (('list', (), _M_omero.cmd._t_RequestList, False, 0),))
    ListRequestsRsp._ice_type = _M_omero.cmd._t_ListRequestsRsp

    _M_omero.cmd.ListRequestsRsp = ListRequestsRsp
    del ListRequestsRsp

    _M_omero.cmd.ListRequestsRspPrx = ListRequestsRspPrx
    del ListRequestsRspPrx

if 'PopStatus' not in _M_omero.cmd.__dict__:
    _M_omero.cmd.PopStatus = Ice.createTempClass()
    class PopStatus(_M_omero.cmd.Request):
        def __init__(self, limit=0, include=None, exclude=None):
            _M_omero.cmd.Request.__init__(self)
            self.limit = limit
            self.include = include
            self.exclude = exclude

        def ice_ids(self, current=None):
            return ('::Ice::Object', '::omero::cmd::PopStatus', '::omero::cmd::Request')

        def ice_id(self, current=None):
            return '::omero::cmd::PopStatus'

        def ice_staticId():
            return '::omero::cmd::PopStatus'
        ice_staticId = staticmethod(ice_staticId)

        def __str__(self):
            return IcePy.stringify(self, _M_omero.cmd._t_PopStatus)

        __repr__ = __str__

    _M_omero.cmd.PopStatusPrx = Ice.createTempClass()
    class PopStatusPrx(_M_omero.cmd.RequestPrx):

        def checkedCast(proxy, facetOrCtx=None, _ctx=None):
            return _M_omero.cmd.PopStatusPrx.ice_checkedCast(proxy, '::omero::cmd::PopStatus', facetOrCtx, _ctx)
        checkedCast = staticmethod(checkedCast)

        def uncheckedCast(proxy, facet=None):
            return _M_omero.cmd.PopStatusPrx.ice_uncheckedCast(proxy, facet)
        uncheckedCast = staticmethod(uncheckedCast)

        def ice_staticId():
            return '::omero::cmd::PopStatus'
        ice_staticId = staticmethod(ice_staticId)

    _M_omero.cmd._t_PopStatusPrx = IcePy.defineProxy('::omero::cmd::PopStatus', PopStatusPrx)

    _M_omero.cmd._t_PopStatus = IcePy.defineClass('::omero::cmd::PopStatus', PopStatus, -1, (), False, False, _M_omero.cmd._t_Request, (), (
        ('limit', (), IcePy._t_int, False, 0),
        ('include', (), _M_omero.cmd._t_StateList, False, 0),
        ('exclude', (), _M_omero.cmd._t_StateList, False, 0)
    ))
    PopStatus._ice_type = _M_omero.cmd._t_PopStatus

    _M_omero.cmd.PopStatus = PopStatus
    del PopStatus

    _M_omero.cmd.PopStatusPrx = PopStatusPrx
    del PopStatusPrx

if 'PopStatusRsp' not in _M_omero.cmd.__dict__:
    _M_omero.cmd.PopStatusRsp = Ice.createTempClass()
    class PopStatusRsp(_M_omero.cmd.OK):
        def __init__(self, list=None):
            _M_omero.cmd.OK.__init__(self)
            self.list = list

        def ice_ids(self, current=None):
            return ('::Ice::Object', '::omero::cmd::OK', '::omero::cmd::PopStatusRsp', '::omero::cmd::Response')

        def ice_id(self, current=None):
            return '::omero::cmd::PopStatusRsp'

        def ice_staticId():
            return '::omero::cmd::PopStatusRsp'
        ice_staticId = staticmethod(ice_staticId)

        def __str__(self):
            return IcePy.stringify(self, _M_omero.cmd._t_PopStatusRsp)

        __repr__ = __str__

    _M_omero.cmd.PopStatusRspPrx = Ice.createTempClass()
    class PopStatusRspPrx(_M_omero.cmd.OKPrx):

        def checkedCast(proxy, facetOrCtx=None, _ctx=None):
            return _M_omero.cmd.PopStatusRspPrx.ice_checkedCast(proxy, '::omero::cmd::PopStatusRsp', facetOrCtx, _ctx)
        checkedCast = staticmethod(checkedCast)

        def uncheckedCast(proxy, facet=None):
            return _M_omero.cmd.PopStatusRspPrx.ice_uncheckedCast(proxy, facet)
        uncheckedCast = staticmethod(uncheckedCast)

        def ice_staticId():
            return '::omero::cmd::PopStatusRsp'
        ice_staticId = staticmethod(ice_staticId)

    _M_omero.cmd._t_PopStatusRspPrx = IcePy.defineProxy('::omero::cmd::PopStatusRsp', PopStatusRspPrx)

    _M_omero.cmd._t_PopStatusRsp = IcePy.declareClass('::omero::cmd::PopStatusRsp')

    _M_omero.cmd._t_PopStatusRsp = IcePy.defineClass('::omero::cmd::PopStatusRsp', PopStatusRsp, -1, (), False, False, _M_omero.cmd._t_OK, (), (('list', (), _M_omero.cmd._t_StatusList, False, 0),))
    PopStatusRsp._ice_type = _M_omero.cmd._t_PopStatusRsp

    _M_omero.cmd.PopStatusRsp = PopStatusRsp
    del PopStatusRsp

    _M_omero.cmd.PopStatusRspPrx = PopStatusRspPrx
    del PopStatusRspPrx

if 'FindHandles' not in _M_omero.cmd.__dict__:
    _M_omero.cmd.FindHandles = Ice.createTempClass()
    class FindHandles(_M_omero.cmd.Request):
        def __init__(self, limit=0, include=None, exclude=None):
            _M_omero.cmd.Request.__init__(self)
            self.limit = limit
            self.include = include
            self.exclude = exclude

        def ice_ids(self, current=None):
            return ('::Ice::Object', '::omero::cmd::FindHandles', '::omero::cmd::Request')

        def ice_id(self, current=None):
            return '::omero::cmd::FindHandles'

        def ice_staticId():
            return '::omero::cmd::FindHandles'
        ice_staticId = staticmethod(ice_staticId)

        def __str__(self):
            return IcePy.stringify(self, _M_omero.cmd._t_FindHandles)

        __repr__ = __str__

    _M_omero.cmd.FindHandlesPrx = Ice.createTempClass()
    class FindHandlesPrx(_M_omero.cmd.RequestPrx):

        def checkedCast(proxy, facetOrCtx=None, _ctx=None):
            return _M_omero.cmd.FindHandlesPrx.ice_checkedCast(proxy, '::omero::cmd::FindHandles', facetOrCtx, _ctx)
        checkedCast = staticmethod(checkedCast)

        def uncheckedCast(proxy, facet=None):
            return _M_omero.cmd.FindHandlesPrx.ice_uncheckedCast(proxy, facet)
        uncheckedCast = staticmethod(uncheckedCast)

        def ice_staticId():
            return '::omero::cmd::FindHandles'
        ice_staticId = staticmethod(ice_staticId)

    _M_omero.cmd._t_FindHandlesPrx = IcePy.defineProxy('::omero::cmd::FindHandles', FindHandlesPrx)

    _M_omero.cmd._t_FindHandles = IcePy.defineClass('::omero::cmd::FindHandles', FindHandles, -1, (), False, False, _M_omero.cmd._t_Request, (), (
        ('limit', (), IcePy._t_int, False, 0),
        ('include', (), _M_omero.cmd._t_StateList, False, 0),
        ('exclude', (), _M_omero.cmd._t_StateList, False, 0)
    ))
    FindHandles._ice_type = _M_omero.cmd._t_FindHandles

    _M_omero.cmd.FindHandles = FindHandles
    del FindHandles

    _M_omero.cmd.FindHandlesPrx = FindHandlesPrx
    del FindHandlesPrx

if 'FindHandlesRsp' not in _M_omero.cmd.__dict__:
    _M_omero.cmd.FindHandlesRsp = Ice.createTempClass()
    class FindHandlesRsp(_M_omero.cmd.OK):
        def __init__(self, handles=None):
            _M_omero.cmd.OK.__init__(self)
            self.handles = handles

        def ice_ids(self, current=None):
            return ('::Ice::Object', '::omero::cmd::FindHandlesRsp', '::omero::cmd::OK', '::omero::cmd::Response')

        def ice_id(self, current=None):
            return '::omero::cmd::FindHandlesRsp'

        def ice_staticId():
            return '::omero::cmd::FindHandlesRsp'
        ice_staticId = staticmethod(ice_staticId)

        def __str__(self):
            return IcePy.stringify(self, _M_omero.cmd._t_FindHandlesRsp)

        __repr__ = __str__

    _M_omero.cmd.FindHandlesRspPrx = Ice.createTempClass()
    class FindHandlesRspPrx(_M_omero.cmd.OKPrx):

        def checkedCast(proxy, facetOrCtx=None, _ctx=None):
            return _M_omero.cmd.FindHandlesRspPrx.ice_checkedCast(proxy, '::omero::cmd::FindHandlesRsp', facetOrCtx, _ctx)
        checkedCast = staticmethod(checkedCast)

        def uncheckedCast(proxy, facet=None):
            return _M_omero.cmd.FindHandlesRspPrx.ice_uncheckedCast(proxy, facet)
        uncheckedCast = staticmethod(uncheckedCast)

        def ice_staticId():
            return '::omero::cmd::FindHandlesRsp'
        ice_staticId = staticmethod(ice_staticId)

    _M_omero.cmd._t_FindHandlesRspPrx = IcePy.defineProxy('::omero::cmd::FindHandlesRsp', FindHandlesRspPrx)

    _M_omero.cmd._t_FindHandlesRsp = IcePy.defineClass('::omero::cmd::FindHandlesRsp', FindHandlesRsp, -1, (), False, False, _M_omero.cmd._t_OK, (), (('handles', (), _M_omero.cmd._t_HandleList, False, 0),))
    FindHandlesRsp._ice_type = _M_omero.cmd._t_FindHandlesRsp

    _M_omero.cmd.FindHandlesRsp = FindHandlesRsp
    del FindHandlesRsp

    _M_omero.cmd.FindHandlesRspPrx = FindHandlesRspPrx
    del FindHandlesRspPrx

if 'Timing' not in _M_omero.cmd.__dict__:
    _M_omero.cmd.Timing = Ice.createTempClass()
    class Timing(_M_omero.cmd.Request):
        """
        Diagnostic command which can be used to see the overhead
        of callbacks. The number of steps and the simulated workload
        can be specified.
        Members:
        steps -- Number of steps that will be run by this command. Value is
        limited by the overall invocation time (5 minutes) as well as
        total number of calls (e.g. 100000)
        millisPerStep -- Number of millis to wait. This value simulates activity on the server.
        Value is limited by the overall invocation time (5 minutes).
        """
        def __init__(self, steps=0, millisPerStep=0):
            _M_omero.cmd.Request.__init__(self)
            self.steps = steps
            self.millisPerStep = millisPerStep

        def ice_ids(self, current=None):
            return ('::Ice::Object', '::omero::cmd::Request', '::omero::cmd::Timing')

        def ice_id(self, current=None):
            return '::omero::cmd::Timing'

        def ice_staticId():
            return '::omero::cmd::Timing'
        ice_staticId = staticmethod(ice_staticId)

        def __str__(self):
            return IcePy.stringify(self, _M_omero.cmd._t_Timing)

        __repr__ = __str__

    _M_omero.cmd.TimingPrx = Ice.createTempClass()
    class TimingPrx(_M_omero.cmd.RequestPrx):

        def checkedCast(proxy, facetOrCtx=None, _ctx=None):
            return _M_omero.cmd.TimingPrx.ice_checkedCast(proxy, '::omero::cmd::Timing', facetOrCtx, _ctx)
        checkedCast = staticmethod(checkedCast)

        def uncheckedCast(proxy, facet=None):
            return _M_omero.cmd.TimingPrx.ice_uncheckedCast(proxy, facet)
        uncheckedCast = staticmethod(uncheckedCast)

        def ice_staticId():
            return '::omero::cmd::Timing'
        ice_staticId = staticmethod(ice_staticId)

    _M_omero.cmd._t_TimingPrx = IcePy.defineProxy('::omero::cmd::Timing', TimingPrx)

    _M_omero.cmd._t_Timing = IcePy.defineClass('::omero::cmd::Timing', Timing, -1, (), False, False, _M_omero.cmd._t_Request, (), (
        ('steps', (), IcePy._t_int, False, 0),
        ('millisPerStep', (), IcePy._t_int, False, 0)
    ))
    Timing._ice_type = _M_omero.cmd._t_Timing

    _M_omero.cmd.Timing = Timing
    del Timing

    _M_omero.cmd.TimingPrx = TimingPrx
    del TimingPrx

# End of module omero.cmd

__name__ = 'omero'

# End of module omero



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# 
# Copyright 2009 Glencoe Software, Inc.  All Rights Reserved.
# Use is subject to license terms supplied in LICENSE.txt
#

"""
OMERO Decorators
"""
from builtins import object
import time
import logging
import traceback

import omero

from functools import wraps

perf_log = logging.getLogger("omero.perf")


def perf(func):
    """ Decorator for (optionally) printing performance statistics """
    def handler(*args, **kwargs):

        # Early Exit. Can't do this in up a level
        # because logging hasn't been configured yet.
        lvl = perf_log.getEffectiveLevel()
        if lvl > logging.DEBUG:
            return func(*args, **kwargs)

        try:
            self = args[0]
            mod = self.__class__.__module__
            cls = self.__class__.__name__
            tag = "%s.%s.%s" % (mod, cls, func.__name__)
        except:
            tag = func.__name__
        start = time.time()
        try:
            rv = func(*args, **kwargs)
            return rv
        finally:
            stop = time.time()
            diff = stop - start
            startMillis = int(start * 1000)
            timeMillis = int(diff * 1000)
            perf_log.debug(
                "start[%d] time[%d] tag[%s]", startMillis, timeMillis, tag)
    handler = wraps(func)(handler)
    return handler


__FORMAT = "%-.120s"
__RESULT = " Rslt: " + __FORMAT
__EXCEPT = " Excp: " + __FORMAT


def remoted(func):
    """
    Decorator for catching any uncaught exception
    and converting it to an InternalException
    """
    log = logging.getLogger("omero.remote")

    def exc_handler(*args, **kwargs):
        try:
            self = args[0]
            log.info(" Meth: %s.%s", self.__class__.__name__, func.__name__)
            rv = func(*args, **kwargs)
            log.info(__RESULT, rv)
            return rv
        except Exception as e:
            log.info(__EXCEPT, e)
            if isinstance(e, omero.ServerError):
                raise
            else:
                log.warn(
                    "%s raised a non-ServerError (%s): %s", func, type(e), e)
                msg = traceback.format_exc()
                raise omero.InternalException(msg, None, "Internal exception")
    exc_handler = wraps(func)(exc_handler)
    return exc_handler


def locked(func):
    """ Decorator for using the self._lock argument of the calling instance """
    def with_lock(*args, **kwargs):
        self = args[0]
        self._lock.acquire()
        try:
            return func(*args, **kwargs)
        finally:
            self._lock.release()
    with_lock = wraps(func)(with_lock)
    return with_lock


class TimeIt (object):

    """
    Decorator to measure the execution time of a function. Assumes that a
    logger global var
    is available and is the logger instance from :meth:`logging.getLogger`.

    @param level: the level to use for logging
    @param name: the name to use when logging, function name is used if None
    """
    logger = logging.getLogger('omero.timeit')

    def __init__(self, level=logging.DEBUG, name=None):
        self._level = level
        self._name = name

    def __call__(self, func):
        def wrapped(*args, **kwargs):
            name = self._name or func.__name__
            self.logger.log(self._level, "timing %s" % (name))
            now = time.time()
            rv = func(*args, **kwargs)
            self.logger.log(self._level, "timed %s: %f" %
                            (name, time.time() - now))
            return rv
        return wrapped


def timeit(func):
    """
    Shortcut version of the :class:`TimeIt` decorator class.
    Logs at logging.DEBUG level.
    """
    def wrapped(*args, **kwargs):
        TimeIt.logger.log(logging.DEBUG, "timing %s" % (func.__name__))
        now = time.time()
        rv = func(*args, **kwargs)
        TimeIt.logger.log(logging.DEBUG, "timed %s: %f" %
                          (func.__name__, time.time() - now))
        return rv
    return TimeIt()(func)


def setsessiongroup(func):
    """
    For BlitzObjectWrapper class derivate functions,
    sets the session group to match the object group.
    """

    def wrapped(self, *args, **kwargs):
        rev = self._conn.setGroupForSession(
            self.getDetails().getGroup().getId())
        try:
            return func(self, *args, **kwargs)
        finally:
            if rev:
                self._conn.revertGroupForSession()
    return wrapped
